# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""test Grizzly Reporter"""
# pylint: disable=protected-access

import os
import sys
import tarfile

import pytest

from FTB.ProgramConfiguration import ProgramConfiguration
from FTB.Signatures.CrashInfo import CrashInfo

from .reporter import FilesystemReporter, FuzzManagerReporter, Report, Reporter, S3FuzzManagerReporter
from .storage import TestCase


def test_report_01():
    """test creating a simple Report"""
    report = Report("no_dir", dict())
    assert report.path == "no_dir"
    assert report.log_aux is None
    assert report.log_err is None
    assert report.log_out is None
    assert report.stack is None
    assert report.preferred is None
    report.cleanup()

def test_report_02(tmp_path):
    """test from_path() with boring logs (no stack)"""
    (tmp_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (tmp_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    report = Report.from_path(str(tmp_path))
    assert report.path == str(tmp_path)
    assert report.log_err.endswith("log_stderr.txt")
    assert report.log_out.endswith("log_stdout.txt")
    assert report.preferred.endswith("log_stderr.txt")
    assert report.log_aux is None
    assert report.stack is None
    assert Report.DEFAULT_MAJOR == report.major
    assert Report.DEFAULT_MINOR == report.minor
    assert report.prefix is not None
    report.cleanup()
    assert not tmp_path.exists()

def test_report_03(tmp_path):
    """test from_path()"""
    (tmp_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (tmp_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    with (tmp_path / "log_asan_blah.txt").open("wb") as log_fp:
        log_fp.write(b"    #0 0xbad000 in foo /file1.c:123:234\n")
        log_fp.write(b"    #1 0x1337dd in bar /file2.c:1806:19")
    report = Report.from_path(str(tmp_path))
    assert report.path == str(tmp_path)
    assert report.log_aux.endswith("log_asan_blah.txt")
    assert report.log_err.endswith("log_stderr.txt")
    assert report.log_out.endswith("log_stdout.txt")
    assert report.preferred.endswith("log_asan_blah.txt")
    assert report.stack is not None
    assert Report.DEFAULT_MAJOR != report.major
    assert Report.DEFAULT_MINOR != report.minor
    assert report.prefix is not None
    report.cleanup()

def test_report_04(tmp_path):
    """test Report.tail()"""
    tmp_file = tmp_path / "file.txt"
    tmp_file.write_bytes(b"blah\ntest\n123\xEF\x00FOO")
    length = tmp_file.stat().st_size
    # no size limit
    with pytest.raises(AssertionError):
        Report.tail(str(tmp_file), 0)
    assert tmp_file.stat().st_size == length
    Report.tail(str(tmp_file), 3)
    log_data = tmp_file.read_bytes()
    assert log_data.startswith(b"[LOG TAILED]\n")
    assert log_data[13:] == b"FOO"

def test_report_05(tmp_path):
    """test Report.select_logs()"""
    with pytest.raises(IOError, match="log_path does not exist"):
        Report.select_logs("missing_path")
    # small log with nothing interesting
    with (tmp_path / "log_asan.txt.1").open("wb") as log_fp:
        log_fp.write(b"SHORT LOG\n")
        log_fp.write(b"filler line")
    # crash on another thread
    with (tmp_path / "log_asan.txt.2").open("wb") as log_fp:
        log_fp.write(b"GOOD LOG\n")
        log_fp.write(b"==70811==ERROR: AddressSanitizer: SEGV on unknown address 0x00000BADF00D")
        log_fp.write(b" (pc 0x7f4c0bb54c67 bp 0x7f4c07bea380 sp 0x7f4c07bea360 T0)\n")  # must be 2nd line
        # pad out to 6 lines
        for l_no in range(4):
            log_fp.write(b"    #%d blah...\n" % l_no)
    # child log that should be ignored (created when parent crashes)
    with (tmp_path / "log_asan.txt.3").open("wb") as log_fp:
        log_fp.write(b"BAD LOG\n")
        log_fp.write(b"==70811==ERROR: AddressSanitizer: SEGV on unknown address 0x000000000000")
        log_fp.write(b" (pc 0x7f4c0bb54c67 bp 0x7f4c07bea380 sp 0x7f4c07bea360 T2)\n")  # must be 2nd line
        # pad out to 6 lines
        for l_no in range(4):
            log_fp.write(b"    #%d blah...\n" % l_no)
    (tmp_path / "log_mindump_blah.txt").write_bytes(b"minidump log")
    (tmp_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (tmp_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    # should be ignored in favor of "GOOD LOG"
    (tmp_path / "log_ffp_worker_blah.txt").write_bytes(b"worker log")
    log_map = Report.select_logs(str(tmp_path))
    assert "GOOD LOG" in (tmp_path / log_map["aux"]).read_text()
    assert "STDERR" in (tmp_path / log_map["stderr"]).read_text()
    assert "STDOUT" in (tmp_path / log_map["stdout"]).read_text()

def test_report_06(tmp_path):
    """test minidump with Report.select_logs()"""
    (tmp_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (tmp_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    with (tmp_path / "log_minidump_01.txt").open("wb") as log_fp:
        log_fp.write(b"GPU|||\n")
        log_fp.write(b"Crash|SIGSEGV|0x0|0\n")
        log_fp.write(b"minidump log\n")
    (tmp_path / "log_ffp_worker_blah.txt").write_bytes(b"worker log")
    log_map = Report.select_logs(str(tmp_path))
    assert (tmp_path / log_map["stderr"]).is_file()
    assert (tmp_path / log_map["stdout"]).is_file()
    assert "minidump log" in (tmp_path / log_map["aux"]).read_text()

def test_report_07(tmp_path):
    """test selecting preferred DUMP_REQUESTED minidump with Report.select_logs()"""
    (tmp_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (tmp_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    with (tmp_path / "log_minidump_01.txt").open("wb") as log_fp:
        log_fp.write(b"GPU|||\n")
        log_fp.write(b"Crash|DUMP_REQUESTED|0x7f9518665d18|0\n")
        log_fp.write(b"0|0|bar.so|sadf|a.cc:739484451a63|3066|0x0\n")
        log_fp.write(b"0|1|gar.so|fdsa|b.cc:739484451a63|1644|0x12\n")
    with (tmp_path / "log_minidump_02.txt").open("wb") as log_fp:
        log_fp.write(b"GPU|||\n")
        log_fp.write(b"Crash|DUMP_REQUESTED|0x7f57ac9e2e14|0\n")
        log_fp.write(b"0|0|foo.so|google_breakpad::ExceptionHandler::WriteMinidump|bar.cc:234|674|0xc\n")
        log_fp.write(b"0|1|foo.so|google_breakpad::ExceptionHandler::WriteMinidump|bar.cc:4a2|645|0x8\n")
    with (tmp_path / "log_minidump_03.txt").open("wb") as log_fp:
        log_fp.write(b"GPU|||\n")
        log_fp.write(b"Crash|DUMP_REQUESTED|0x7f9518665d18|0\n")
        log_fp.write(b"0|0|bar.so|sadf|a.cc:1234|3066|0x0\n")
        log_fp.write(b"0|1|gar.so|fdsa|b.cc:4323|1644|0x12\n")
    log_map = Report.select_logs(str(tmp_path))
    assert (tmp_path / log_map["stderr"]).is_file()
    assert (tmp_path / log_map["stdout"]).is_file()
    assert "google_breakpad::ExceptionHandler::WriteMinidump" in (tmp_path / log_map["aux"]).read_text()

def test_report_08(tmp_path):
    """test selecting worker logs with Report.select_logs()"""
    (tmp_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (tmp_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    (tmp_path / "log_ffp_worker_blah.txt").write_bytes(b"worker log")
    log_map = Report.select_logs(str(tmp_path))
    assert (tmp_path / log_map["stderr"]).is_file()
    assert (tmp_path / log_map["stdout"]).is_file()
    assert "worker log" in (tmp_path / log_map["aux"]).read_text()

def test_report_09(tmp_path):
    """test prioritizing *San logs with Report.select_logs()"""
    # crash
    with (tmp_path / "log_asan.txt.1").open("wb") as log_fp:
        log_fp.write(b"GOOD LOG\n")
        log_fp.write(b"==1942==ERROR: AddressSanitizer: heap-use-after-free on ... blah\n")  # must be 2nd line
        # pad out to 6 lines
        for l_no in range(4):
            log_fp.write(b"    #%d blah...\n" % l_no)
    # crash missing trace
    with (tmp_path / "log_asan.txt.2").open("wb") as log_fp:
        log_fp.write(b"BAD LOG\n")
        log_fp.write(b"==1984==ERROR: AddressSanitizer: SEGV on ... blah\n")  # must be 2nd line
        log_fp.write(b"missing trace...\n")
    # child log that should be ignored (created when parent crashes)
    with (tmp_path / "log_asan.txt.3").open("wb") as log_fp:
        log_fp.write(b"BAD LOG\n")
        log_fp.write(b"==1184==ERROR: AddressSanitizer: BUS on ... blah\n")  # must be 2nd line
        # pad out to 6 lines
        for l_no in range(4):
            log_fp.write(b"    #%d blah...\n" % l_no)
    with (tmp_path / "log_asan.txt.4").open("wb") as log_fp:
        log_fp.write(b"BAD LOG\n")
        log_fp.write(b"==9482==ERROR: AddressSanitizer: stack-overflow on ...\n")  # must be 2nd line
        # pad out to 6 lines
        for l_no in range(4):
            log_fp.write(b"    #%d blah...\n" % l_no)
    with (tmp_path / "log_asan.txt.5").open("wb") as log_fp:
        log_fp.write(b"BAD LOG\n")
        log_fp.write(b"ERROR: Failed to mmap\n")  # must be 2nd line
    log_map = Report.select_logs(str(tmp_path))
    assert "GOOD LOG" in (tmp_path / log_map["aux"]).read_text()

def test_report_10(tmp_path):
    """test Report size_limit"""
    (tmp_path / "log_stderr.txt").write_bytes(b"STDERR log\n" * 200)
    (tmp_path / "log_stdout.txt").write_bytes(b"STDOUT log\n" * 200)
    (tmp_path / "unrelated.txt").write_bytes(b"nothing burger\n" * 200)
    (tmp_path / "rr-trace").mkdir()
    size_limit = len("STDERR log\n")
    report = Report.from_path(str(tmp_path), size_limit=size_limit)
    assert report.path == str(tmp_path)
    assert report.log_err.endswith("log_stderr.txt")
    assert report.log_out.endswith("log_stdout.txt")
    assert report.preferred.endswith("log_stderr.txt")
    assert report.log_aux is None
    assert report.stack is None
    size_limit += len("[LOG TAILED]\n")
    assert os.stat(os.path.join(report.path, report.log_err)).st_size == size_limit
    assert os.stat(os.path.join(report.path, report.log_out)).st_size == size_limit
    assert os.stat(os.path.join(report.path, "unrelated.txt")).st_size == size_limit
    report.cleanup()
    assert not tmp_path.is_dir()

def test_report_11(tmp_path):
    """test selecting Valgrind logs with Report.select_logs()"""
    (tmp_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (tmp_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    (tmp_path / "log_valgrind.txt").write_bytes(b"valgrind log")
    log_map = Report.select_logs(str(tmp_path))
    assert (tmp_path / log_map["stderr"]).is_file()
    assert (tmp_path / log_map["stdout"]).is_file()
    assert "valgrind log" in (tmp_path / log_map["aux"]).read_text()

def test_report_12(tmp_path):
    """test Report.crash_info()"""
    (tmp_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (tmp_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    with (tmp_path / "log_asan_blah.txt").open("wb") as log_fp:
        log_fp.write(b"    #0 0xbad000 in foo /file1.c:123:234\n")
        log_fp.write(b"    #1 0x1337dd in bar /file2.c:1806:19")
    # no binary.fuzzmanagerconf
    report = Report.from_path(str(tmp_path))
    assert report._crash_info is None
    assert report.crash_info("fake_bin") is not None
    assert report._crash_info is not None
    # with binary.fuzzmanagerconf
    with (tmp_path / "fake_bin.fuzzmanagerconf").open("wb") as conf:
        conf.write(b"[Main]\n")
        conf.write(b"platform = x86-64\n")
        conf.write(b"product = mozilla-central\n")
        conf.write(b"os = linux\n")
    report = Report.from_path(str(tmp_path))
    assert report._crash_info is None
    assert report.crash_info(str(tmp_path / "fake_bin")) is not None
    assert report._crash_info is not None

def test_report_13(mocker, tmp_path):
    """test Report.crash_signature() and Report.crash_hash()"""
    mocker.patch("grizzly.common.reporter.ProgramConfiguration", autospec=True)
    (tmp_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (tmp_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    with (tmp_path / "log_asan_blah.txt").open("wb") as log_fp:
        log_fp.write(b"==1==ERROR: AddressSanitizer: SEGV on unknown address 0x0 (pc 0x0 bp 0x0 sp 0x0 T0)\n")
        log_fp.write(b"    #0 0xbad000 in foo /file1.c:123:234\n")
        log_fp.write(b"    #1 0x1337dd in bar /file2.c:1806:19")
    report = Report.from_path(str(tmp_path))
    assert report._crash_info is None
    info = report.crash_info("fake_bin")
    sig = Report.crash_signature(info)
    assert sig.symptoms
    short_sig = info.createShortSignature()
    assert short_sig == "[@ foo]"
    assert Report.crash_hash(info)

def test_report_14(mocker):
    """test Report.crash_signature_max_frames()"""
    info = mocker.Mock(spec=CrashInfo)
    info.backtrace = ("blah")
    assert Report.crash_signature_max_frames(info) == 8
    info.backtrace = ("std::panicking::rust_panic", "std::panicking::rust_panic_with_hook")
    assert Report.crash_signature_max_frames(info) == 14

def test_reporter_01(mocker, tmp_path):
    """test creating a simple Reporter"""
    class SimpleReporter(Reporter):
        def _process_report(self, report):
            pass
        def _reset(self):
            pass
        def _submit_report(self, report, test_cases):
            pass
    reporter = SimpleReporter()
    with pytest.raises(AssertionError, match="Either 'log_path' or 'report' must be specified!"):
        reporter.submit([])
    with pytest.raises(IOError, match="No such directory 'fake_dir'"):
        reporter.submit([], log_path="fake_dir")
    with pytest.raises(IOError, match="No logs found in"):
        reporter.submit([], log_path=str(tmp_path))
    with pytest.raises(AssertionError, match="Only 'log_path' or 'report' can be specified!"):
        reporter.submit([], log_path=str(tmp_path), report=mocker.Mock())
    # submit a report
    reporter.submit([], report=mocker.Mock(spec=Report))

def test_filesystem_reporter_01(tmp_path):
    """test FilesystemReporter without testcases"""
    log_path = tmp_path / "logs"
    log_path.mkdir()
    (log_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (log_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    with (log_path / "log_asan_blah.txt").open("wb") as log_fp:
        log_fp.write(b"    #0 0xbad000 in foo /file1.c:123:234\n")
        log_fp.write(b"    #1 0x1337dd in bar /file2.c:1806:19")
    report_path = tmp_path / "reports"
    report_path.mkdir()
    reporter = FilesystemReporter(report_path=str(report_path))
    reporter.submit([], log_path=str(log_path))
    buckets = [x for x in report_path.iterdir()]
    # check major bucket
    assert len(buckets) == 1
    assert buckets[0].is_dir()
    # check log path exists
    log_dirs = [x for x in buckets[0].iterdir()]
    assert len(log_dirs) == 1
    assert log_dirs[0].is_dir()
    assert "_logs" in str(log_dirs[0])

def test_filesystem_reporter_02(tmp_path, mocker):
    """test FilesystemReporter with testcases"""
    log_path = tmp_path / "logs"
    log_path.mkdir()
    (log_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (log_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    with (log_path / "log_asan_blah.txt").open("wb") as log_fp:
        log_fp.write(b"    #0 0xbad000 in foo /file1.c:123:234\n")
        log_fp.write(b"    #1 0x1337dd in bar /file2.c:1806:19")
    testcases = list()
    for _ in range(10):
        testcases.append(mocker.Mock(spec=TestCase))
    report_path = tmp_path / "reports"
    assert not report_path.exists()
    reporter = FilesystemReporter(report_path=str(report_path))
    reporter.submit(testcases, log_path=str(log_path))
    assert not log_path.exists()
    assert report_path.exists()
    assert len(tuple(report_path.glob("*"))) == 1
    for tstc in testcases:
        assert tstc.dump.call_count == 1
    # call report a 2nd time
    log_path.mkdir()
    (log_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (log_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    testcases = list()
    for _ in range(2):
        testcases.append(mocker.Mock(spec=TestCase))
    reporter.submit(testcases, log_path=str(log_path))
    for tstc in testcases:
        assert tstc.dump.call_count == 1
    assert len(tuple(report_path.glob("*"))) == 2
    assert len(tuple(report_path.glob("NO_STACK"))) == 1

def test_filesystem_reporter_03(tmp_path):
    """test FilesystemReporter disk space failsafe"""
    log_path = tmp_path / "logs"
    log_path.mkdir()
    (log_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (log_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    report_path = tmp_path / "reports"
    report_path.mkdir()
    reporter = FilesystemReporter(report_path=str(report_path))
    reporter.DISK_SPACE_ABORT = 2 ** 50
    with pytest.raises(RuntimeError) as exc:
        reporter.submit([], log_path=str(log_path))
    assert "Running low on disk space" in str(exc.value)

def test_filesystem_reporter_04(mocker, tmp_path):
    """test FilesystemReporter w/o major bucket"""
    report = mocker.Mock(spec=Report)
    report_path = (tmp_path / "report")
    report_path.mkdir()
    report.path = str(report_path)
    report.prefix = "0000_2020_01_01"
    reporter = FilesystemReporter(report_path=str(tmp_path), major_bucket=False)
    reporter.submit([], report=report)
    assert not report_path.is_dir()
    assert not report.major.call_count

def test_fuzzmanager_reporter_01(tmp_path, mocker):
    """test FuzzManagerReporter.sanity_check()"""
    fake_reporter = mocker.patch("grizzly.common.reporter.ProgramConfiguration")
    fake_reporter.fromBinary.return_value = mocker.Mock(spec=ProgramConfiguration)
    # missing global FM config file
    FuzzManagerReporter.FM_CONFIG = "no_file"
    with pytest.raises(IOError, match="Missing: no_file"):
        FuzzManagerReporter.sanity_check("fake")
    # missing binary FM config file
    fake_fmc = tmp_path / ".fuzzmanagerconf"
    fake_fmc.touch()
    fake_bin = tmp_path / "bin"
    fake_bin.touch()
    FuzzManagerReporter.FM_CONFIG = str(fake_fmc)
    with pytest.raises(IOError, match="bin.fuzzmanagerconf"):
        FuzzManagerReporter.sanity_check(str(fake_bin))
    # success
    (tmp_path / "bin.fuzzmanagerconf").touch()
    FuzzManagerReporter.sanity_check(str(fake_bin))
    assert fake_reporter.fromBinary.call_count == 1

def test_fuzzmanager_reporter_02(tmp_path):
    """test FuzzManagerReporter.submit() empty path"""
    reporter = FuzzManagerReporter("fake_bin")
    report_path = tmp_path / "report"
    report_path.mkdir()
    with pytest.raises(IOError) as exc:
        reporter.submit([], log_path=str(report_path))
    assert "No logs found in" in str(exc.value)

def test_fuzzmanager_reporter_03(tmp_path, mocker):
    """test FuzzManagerReporter.submit()"""
    fake_crashinfo = mocker.patch("grizzly.common.reporter.CrashInfo", autospec=True)
    fake_crashinfo.fromRawCrashData.return_value.createShortSignature.return_value = "test [@ test]"
    fake_collector = mocker.patch("grizzly.common.reporter.Collector", autospec=True)
    fake_collector.return_value.search.return_value = (None, None)
    fake_collector.return_value.generate.return_value = str(tmp_path / "fake_sig_file")
    log_path = tmp_path / "log_path"
    log_path.mkdir()
    (log_path / "log_ffp_worker_blah.txt").touch()
    (log_path / "log_stderr.txt").touch()
    (log_path / "log_stdout.txt").touch()
    report = Report.from_path(str(log_path))
    fake_test = mocker.Mock(spec=TestCase)
    fake_test.adapter_name = "adapter"
    fake_test.input_fname = "input"
    fake_test.env_vars = {"TEST": "1"}
    reporter = FuzzManagerReporter(str("fake_bin"))
    reporter.submit([fake_test], report=report)
    assert not log_path.is_dir()
    assert fake_test.dump.call_count == 1
    assert fake_collector.return_value.submit.call_count == 1

def test_fuzzmanager_reporter_04(tmp_path, mocker):
    """test FuzzManagerReporter.submit() hit frequent crash"""
    mocker.patch("grizzly.common.reporter.CrashInfo", autospec=True)
    fake_collector = mocker.patch("grizzly.common.reporter.Collector", autospec=True)
    fake_collector.return_value.search.return_value = (None, {"frequent": True, "shortDescription": "[@ test]"})
    reporter = FuzzManagerReporter("fake_bin")
    log_path = tmp_path / "log_path"
    log_path.mkdir()
    (log_path / "log_stderr.txt").touch()
    (log_path / "log_stdout.txt").touch()
    reporter.submit([], log_path=str(log_path))
    fake_collector.return_value.submit.assert_not_called()

def test_fuzzmanager_reporter_05(tmp_path, mocker):
    """test FuzzManagerReporter.submit() hit existing crash"""
    mocker.patch("grizzly.common.reporter.CrashInfo", autospec=True)
    fake_collector = mocker.patch("grizzly.common.reporter.Collector", autospec=True)
    fake_collector.return_value.search.return_value = (
        None, {"bug__id":1, "frequent": False, "shortDescription": "[@ test]"})
    reporter = FuzzManagerReporter("fake_bin")
    log_path = tmp_path / "log_path"
    log_path.mkdir()
    (log_path / "log_stderr.txt").touch()
    (log_path / "log_stdout.txt").touch()
    reporter._ignored = lambda x: True
    reporter.submit([], log_path=str(log_path))
    fake_collector.return_value.submit.assert_not_called()

def test_fuzzmanager_reporter_06(tmp_path, mocker):
    """test FuzzManagerReporter.submit() no signature"""
    mocker.patch("grizzly.common.reporter.CrashInfo", autospec=True)
    fake_collector = mocker.patch("grizzly.common.reporter.Collector", autospec=True)
    fake_collector.return_value.search.return_value = (None, None)
    fake_collector.return_value.generate.return_value = None
    reporter = FuzzManagerReporter("fake_bin")
    log_path = tmp_path / "log_path"
    log_path.mkdir()
    (log_path / "log_stderr.txt").touch()
    (log_path / "log_stdout.txt").touch()
    with pytest.raises(RuntimeError) as exc:
        reporter.submit([], log_path=str(log_path))
    assert "Failed to create FM signature" in str(exc.value)
    fake_collector.return_value.submit.assert_not_called()
    # test ignore unsymbolized crash
    reporter._ignored = lambda x: True
    reporter.submit([], log_path=str(log_path))
    fake_collector.return_value.submit.assert_not_called()

def test_s3fuzzmanager_reporter_01(tmp_path, mocker):
    """test S3FuzzManagerReporter.sanity_check()"""
    mocker.patch("grizzly.common.reporter.FuzzManagerReporter", autospec=True)
    fake_bin = tmp_path / "bin"
    with pytest.raises(EnvironmentError) as exc:
        S3FuzzManagerReporter.sanity_check(str(fake_bin))
    assert "'GRZ_S3_BUCKET' is not set in environment" in str(exc.value)
    pytest.importorskip("boto3")
    os.environ["GRZ_S3_BUCKET"] = "test"
    try:
        S3FuzzManagerReporter.sanity_check(str(fake_bin))
    finally:
        os.environ.pop("GRZ_S3_BUCKET", None)

def test_s3fuzzmanager_reporter_02(tmp_path, mocker):
    """test S3FuzzManagerReporter._process_report()"""
    pytest.importorskip("boto3")
    pytest.importorskip("botocore")
    fake_boto3 = mocker.patch("grizzly.common.reporter.boto3", autospec=True)

    fake_report = mocker.Mock(spec=Report)
    fake_report.path = "no-path"
    reporter = S3FuzzManagerReporter("fake_bin")
    # test will missing rr-trace
    assert reporter._process_report(fake_report) is None
    assert not reporter._extra_metadata

    # test will exiting rr-trace
    trace_dir = tmp_path / "rr-traces" / "latest-trace"
    trace_dir.mkdir(parents=True)
    fake_report.minor = "1234abcd"
    fake_report.path = str(tmp_path)
    os.environ["GRZ_S3_BUCKET"] = "test"
    try:
        reporter._process_report(fake_report)
    finally:
        os.environ.pop("GRZ_S3_BUCKET", None)
    assert not tuple(tmp_path.glob("*"))
    assert "rr-trace" in reporter._extra_metadata
    assert fake_report.minor in reporter._extra_metadata["rr-trace"]
    fake_boto3.resource.return_value.meta.client.upload_file.assert_not_called()

    # test with new rr-trace
    reporter._extra_metadata.clear()
    trace_dir.mkdir(parents=True)
    (trace_dir / "trace-file").touch()
    class FakeClientError(Exception):
        def __init__(self, message, response):
            super(FakeClientError, self).__init__(message)
            self.response = response
    fake_botocore = mocker.patch("grizzly.common.reporter.botocore", autospec=True)
    fake_botocore.exceptions.ClientError = FakeClientError
    fake_boto3.resource.return_value.Object.side_effect = FakeClientError("test", {"Error": {"Code": "404"}})
    os.environ["GRZ_S3_BUCKET"] = "test"
    try:
        reporter._process_report(fake_report)
    finally:
        os.environ.pop("GRZ_S3_BUCKET", None)
    assert not tuple(tmp_path.glob("*"))
    assert "rr-trace" in reporter._extra_metadata
    assert fake_report.minor in reporter._extra_metadata["rr-trace"]
    assert fake_boto3.resource.return_value.meta.client.upload_file.call_count == 1

@pytest.mark.skipif(not sys.platform.startswith("linux"), reason="RR only supported on Linux")
def test_s3fuzzmanager_reporter_03(tmp_path):
    """test S3FuzzManagerReporter.compress_rr_trace()"""
    # create fake trace
    src = tmp_path / "rr-traces" / "echo-0"
    src.mkdir(parents=True)
    (src / "fail_file").touch()
    src = tmp_path / "rr-traces" / "echo-1"
    src.mkdir()
    (src / "cloned_data_5799_1").touch()
    (src / "data").write_bytes(b"test_data")
    (src / "events").write_bytes(b"foo")
    (src / "mmap").write_bytes(b"bar")
    (src / "tasks").write_bytes(b"foo")
    (src / "version").write_bytes(b"123")
    (tmp_path / "rr-traces" / "latest-trace").symlink_to(str(src), target_is_directory=True)
    src = tmp_path / "rr-traces"
    dest = tmp_path / "dest"
    dest.mkdir()
    S3FuzzManagerReporter.compress_rr_trace(str(src), str(dest))
    assert not src.is_dir()
    assert (dest / "rr.tar.bz2").is_file()
    with tarfile.open(str(dest / "rr.tar.bz2"), "r:bz2") as arc_fp:
        entries = arc_fp.getnames()
    assert "echo-1" in entries
    assert "echo-0" not in entries
    assert "latest-trace" not in entries

# TODO: fill out tests for FuzzManagerReporter and S3FuzzManagerReporter



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access

from .iomanager import IOManager
from .storage import TestFile


def test_iomanager_01():
    """test a simple IOManager"""
    with IOManager() as iom:
        assert iom.server_map is not None
        assert not iom._environ_files
        assert iom._generated == 0
        assert iom._report_size == 1

def test_iomanager_02():
    """test IOManager.page_name()"""
    with IOManager() as iom:
        assert iom.page_name() != iom.page_name(offset=1)
        next_page = iom.page_name(offset=1)
        iom._generated += 1
        assert iom.page_name() == next_page

def test_iomanager_03(mocker, tmp_path):
    """test IOManager._add_suppressions()"""
    mocker.patch.dict("grizzly.common.iomanager.environ", values={})
    with IOManager() as iom:
        assert not iom._environ_files
        supp_file = tmp_path / "supp_file.txt"
        supp_file.touch()
        mocker.patch.dict(
            "grizzly.common.iomanager.environ",
            values={
                "ASAN_OPTIONS": "blah=1:suppressions='%s':foo=2" % (str(supp_file),),
                "DEBUG": "1",
                "LSAN_OPTIONS": "nothing=1",
                "JUNK": "test"})
        iom._add_suppressions()
        assert "asan.supp" in (x.file_name for x in iom._environ_files)

def test_iomanager_04():
    """test IOManager.create_testcase()"""
    with IOManager() as iom:
        assert iom._generated == 0
        assert iom._report_size == 1
        assert not iom.tests
        assert not iom.server_map.dynamic
        assert not iom.server_map.include
        assert not iom.server_map.redirect
        iom._tracked_env = {"TEST": "1"}
        iom._environ_files = [TestFile.from_data(b"data", "e.txt")]
        # without a harness, no input files
        tcase = iom.create_testcase("test-adapter")
        assert tcase is not None
        assert iom._generated == 1
        assert len(iom.tests) == 1
        assert not any(tcase.optional)
        assert "grz_current_test" in iom.server_map.redirect
        assert iom.server_map.redirect["grz_current_test"].target == tcase.landing_page
        assert "grz_next_test" in iom.server_map.redirect
        assert "grz_harness" not in iom.server_map.dynamic
        # with a harness
        iom.harness = b"harness-data"
        tcase = iom.create_testcase("test-adapter")
        assert tcase is not None
        assert len(iom.tests) == 1
        assert iom._generated == 2
        assert "grz_current_test" in iom.server_map.redirect
        assert iom.server_map.redirect["grz_current_test"].target == tcase.landing_page
        assert "grz_next_test" in iom.server_map.redirect
        assert "grz_harness" in iom.server_map.dynamic

def test_iomanager_05(mocker):
    """test IOManager.tracked_environ()"""
    mocker.patch.dict("grizzly.common.iomanager.environ", values={})
    assert not IOManager.tracked_environ()
    mocker.patch.dict(
        "grizzly.common.iomanager.environ",
        values={
            "ASAN_OPTIONS": "blah='z:/a':detect_leaks=1:foo=2",
            "LSAN_OPTIONS": "detect_leaks='x:\\a.1':a=1",
            # should be added since it is in IOManager.TRACKED_ENVVARS
            "MOZ_CHAOSMODE": "1",
            # this should be skipped because it uses the FFPuppet debug
            "XPCOM_DEBUG_BREAK": "warn",
            "TEST_BAD": "FAIL"},
        clear=True)
    tracked = IOManager.tracked_environ()
    assert "TEST_BAD" not in tracked
    assert "XPCOM_DEBUG_BREAK" not in tracked
    assert "ASAN_OPTIONS" in tracked
    assert "MOZ_CHAOSMODE" in tracked
    assert tracked["ASAN_OPTIONS"] == "detect_leaks=1"
    assert "LSAN_OPTIONS" in tracked
    assert tracked["LSAN_OPTIONS"] == "detect_leaks='x:\\a.1'"
    mocker.patch.dict(
        "grizzly.common.iomanager.environ",
        values={"ASAN_OPTIONS": "ignored=x"},
        clear=True)
    assert not IOManager.tracked_environ()



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
from __future__ import unicode_literals
import json
import os.path
import zipfile
import pytest
from grizzly.reduce import exceptions, strategies
from grizzly.common import FuzzManagerReporter
from .test_common import BaseFakeReporter, TestReductionJob
from .test_common import TestReductionJobAlt, TestReductionJobKeepHarness, TestReductionJobSemiReliable


def test_config_testcase_0(tmp_path):
    """empty directory fails config_testcase"""
    with TestReductionJob(tmp_path, create_binary=False) as job:
        with pytest.raises(exceptions.NoTestcaseError) as exc:
            job.config_testcase(str(tmp_path))
        assert "No testcase recognized" in str(exc.value)
        assert job.result_code == FuzzManagerReporter.QUAL_NO_TESTCASE


def test_config_testcase_1(tmp_path):
    """non-zip file fails config_testcase"""
    with TestReductionJob(tmp_path, create_binary=False) as job:
        tmp_file = tmp_path / "test.txt"
        tmp_file.touch()
        with pytest.raises(exceptions.ReducerError) as exc:
            job.config_testcase(str(tmp_file))
        assert "Testcase must be zip, html, or directory" in str(exc.value)
        tmp_file.unlink()
        assert job.result_code == FuzzManagerReporter.QUAL_REDUCER_ERROR


def test_config_testcase_2(tmp_path):
    """config_testcase can't be called twice"""
    with TestReductionJob(tmp_path, create_binary=False) as job:
        with pytest.raises(exceptions.ReducerError) as exc:
            job.config_testcase(str(tmp_path))
        with pytest.raises(exceptions.ReducerError) as exc:
            job.config_testcase(str(tmp_path))
        assert "Testcase already configured?" in str(exc.value)
        assert job.result_code == FuzzManagerReporter.QUAL_REDUCER_ERROR


def test_config_testcase_3(tmp_path):
    """bad zip file fails config_testcase"""
    with TestReductionJob(tmp_path, create_binary=False) as job:
        test_zip = tmp_path / "test.zip"
        test_zip.touch()
        with pytest.raises(exceptions.CorruptTestcaseError):
            job.config_testcase(str(test_zip))
        assert job.result_code == FuzzManagerReporter.QUAL_REDUCER_ERROR


def test_config_testcase_4(tmp_path):
    """missing landing page causes failure"""
    with TestReductionJob(tmp_path, create_binary=False) as job:
        (tmp_path / "test_info.json").write_text("{}")
        with pytest.raises(exceptions.ReducerError) as exc:
            job.config_testcase(str(tmp_path))
        assert "Could not find landing page" in str(exc.value)
        assert job.result_code == FuzzManagerReporter.QUAL_REDUCER_ERROR


def test_config_testcase_4_legacy(tmp_path):
    """missing landing page causes failure"""
    # TODO: remove this test
    with TestReductionJob(tmp_path, create_binary=False) as job:
        (tmp_path / "test_info.txt").touch()
        with pytest.raises(exceptions.ReducerError) as exc:
            job.config_testcase(str(tmp_path))
        assert "Could not find landing page" in str(exc.value)
        assert job.result_code == FuzzManagerReporter.QUAL_REDUCER_ERROR


def test_config_testcase_5(tmp_path):
    """missing landing page causes failure"""
    with TestReductionJob(tmp_path, create_binary=False) as job:
        (tmp_path / "test_info.json").write_text("{\"target\":\"\",\"env\":{}}")
        with pytest.raises(exceptions.ReducerError) as exc:
            job.config_testcase(str(tmp_path))
        assert "does not exist" in str(exc.value)
        assert job.result_code == FuzzManagerReporter.QUAL_REDUCER_ERROR


def test_config_testcase_5_legacy(tmp_path):
    """missing landing page causes failure"""
    # TODO: remove this test
    with TestReductionJob(tmp_path, create_binary=False) as job:
        (tmp_path / "test_info.txt").write_text("landing page: ")
        with pytest.raises(exceptions.ReducerError) as exc:
            job.config_testcase(str(tmp_path))
        assert "does not exist" in str(exc.value)
        assert job.result_code == FuzzManagerReporter.QUAL_REDUCER_ERROR


def test_config_testcase_6(tmp_path):
    """single testcase is loaded ok"""
    with TestReductionJob(tmp_path, create_binary=False) as job:
        (tmp_path / "test_info.json").write_text("{\"target\":\"test.html\",\"env\":{}}")
        (tmp_path / "test.html").write_text("hello")
        job.config_testcase(str(tmp_path))
        assert job._testcase == os.path.join(job._tcroot, "test.html")
        with open(job._testcase) as tc_fp:
            assert tc_fp.read() == "hello"
        assert job.result_code is None


def test_config_testcase_6_legacy(tmp_path):
    """single testcase is loaded ok"""
    # TODO: remove this test
    with TestReductionJob(tmp_path, create_binary=False) as job:
        (tmp_path / "test_info.txt").write_text("landing page: test.html")
        (tmp_path / "test.html").write_text("hello")
        job.config_testcase(str(tmp_path))
        assert job._testcase == os.path.join(job._tcroot, "test.html")
        with open(job._testcase) as tc_fp:
            assert tc_fp.read() == "hello"
        assert job.result_code is None


def test_config_testcase_7(tmp_path):
    """single testcase in numbered subdir is loaded ok"""
    with TestReductionJob(tmp_path, create_binary=False) as job:
        (tmp_path / "-0").mkdir()
        (tmp_path / "-0" / "test_info.json").write_text("{\"target\":\"test.html\",\"env\":{}}")
        (tmp_path / "-0" / "test.html").write_text("hello")
        job.config_testcase(str(tmp_path))
        assert job._testcase == os.path.join(job._tcroot, "-0", "test.html")
        with open(job._testcase) as tc_fp:
            assert tc_fp.read() == "hello"
        assert job.result_code is None


def test_config_testcase_7_legacy(tmp_path):
    """single testcase in numbered subdir is loaded ok"""
    # TODO: remove this test
    with TestReductionJob(tmp_path, create_binary=False) as job:
        (tmp_path / "-0").mkdir()
        (tmp_path / "-0" / "test_info.txt").write_text("landing page: test.html")
        (tmp_path / "-0" / "test.html").write_text("hello")
        job.config_testcase(str(tmp_path))
        assert job._testcase == os.path.join(job._tcroot, "-0", "test.html")
        with open(job._testcase) as tc_fp:
            assert tc_fp.read() == "hello"
        assert job.result_code is None


def test_config_testcase_8(tmp_path):
    """multiple testcase in numbered subdir creates a harness"""
    with TestReductionJob(tmp_path, create_binary=False) as job:
        for subdir in ("-0", "-1"):
            (tmp_path / subdir).mkdir()
            (tmp_path / subdir / "test_info.json").write_text("{\"target\":\"test.html\",\"env\":{}}")
            (tmp_path / subdir / "test.html").write_text("hello")
        job.config_testcase(str(tmp_path))
        assert job._testcase.startswith(os.path.join(job._tcroot, "harness_"))
        assert job._testcase.endswith(".html")
        for subdir in ("-0", "-1"):
            with open(os.path.join(job._tcroot, subdir, "test.html")) as tc_fp:
                assert tc_fp.read() == "hello"
        with open(job._testcase) as tc_fp:
            harness = tc_fp.read()
        loc0 = harness.index("'/-0/test.html',")
        loc1 = harness.index("'/-1/test.html',")
        assert loc1 < loc0, "testcases should occur in harness in descending order"
        assert job.result_code is None


def test_config_testcase_8_legacy(tmp_path):
    """multiple testcase in numbered subdir creates a harness"""
    # TODO: remove this test
    with TestReductionJob(tmp_path, create_binary=False) as job:
        for subdir in ("-0", "-1"):
            (tmp_path / subdir).mkdir()
            (tmp_path / subdir / "test_info.txt").write_text("landing page: test.html")
            (tmp_path / subdir / "test.html").write_text("hello")
        job.config_testcase(str(tmp_path))
        assert job._testcase.startswith(os.path.join(job._tcroot, "harness_"))
        assert job._testcase.endswith(".html")
        for subdir in ("-0", "-1"):
            with open(os.path.join(job._tcroot, subdir, "test.html")) as tc_fp:
                assert tc_fp.read() == "hello"
        with open(job._testcase) as tc_fp:
            harness = tc_fp.read()
        loc0 = harness.index("'/-0/test.html',")
        loc1 = harness.index("'/-1/test.html',")
        assert loc1 < loc0, "testcases should occur in harness in descending order"
        assert job.result_code is None


def test_config_testcase_9(tmp_path):
    """single testcase is loaded from zip ok"""
    with TestReductionJob(tmp_path, create_binary=False) as job:
        (tmp_path / "test_info.json").write_text("{\"target\":\"test.html\",\"env\":{}}")
        (tmp_path / "test.html").write_text("hello")
        with zipfile.ZipFile(str(tmp_path / "test.zip"), "w") as zip_fp:
            zip_fp.write(str(tmp_path / "test_info.json"), "test_info.json")
            zip_fp.write(str(tmp_path / "test.html"), "test.html")
        job.config_testcase(str(tmp_path / "test.zip"))
        assert job._testcase == os.path.join(job._tcroot, "test.html")
        with open(job._testcase) as tc_fp:
            assert tc_fp.read() == "hello"
        assert job.result_code is None


def test_config_testcase_9_legacy(tmp_path):
    """single testcase is loaded from zip ok"""
    # TODO: remove this test
    with TestReductionJob(tmp_path, create_binary=False) as job:
        (tmp_path / "test_info.txt").write_text("landing page: test.html")
        (tmp_path / "test.html").write_text("hello")
        with zipfile.ZipFile(str(tmp_path / "test.zip"), "w") as zip_fp:
            zip_fp.write(str(tmp_path / "test_info.txt"), "test_info.txt")
            zip_fp.write(str(tmp_path / "test.html"), "test.html")
        job.config_testcase(str(tmp_path / "test.zip"))
        assert job._testcase == os.path.join(job._tcroot, "test.html")
        with open(job._testcase) as tc_fp:
            assert tc_fp.read() == "hello"
        assert job.result_code is None


def test_config_testcase_10(tmp_path):
    """prefs from testcase are used and take precedence over target prefs"""
    with TestReductionJob(tmp_path, create_binary=False) as job:
        (tmp_path / "orig_prefs.js").write_text("orig prefs")
        job.target.prefs = str(tmp_path / "orig_prefs.js")
        (tmp_path / "test_info.json").write_text("{\"target\":\"test.html\",\"env\":{}}")
        (tmp_path / "test.html").write_text("hello")
        (tmp_path / "prefs.js").write_text("some prefs")
        job.config_testcase(str(tmp_path))
        assert os.path.normpath(job.target.prefs) \
            != os.path.normpath(str(tmp_path / "prefs.js"))
        with open(job.target.prefs) as prefs_fp:
            assert prefs_fp.read() == "some prefs"
        assert job.result_code is None


def test_config_testcase_10_legacy(tmp_path):
    """prefs from testcase are used and take precedence over target prefs"""
    # TODO: remove this test
    with TestReductionJob(tmp_path, create_binary=False) as job:
        (tmp_path / "orig_prefs.js").write_text("orig prefs")
        job.target.prefs = str(tmp_path / "orig_prefs.js")
        (tmp_path / "test_info.txt").write_text("landing page: test.html")
        (tmp_path / "test.html").write_text("hello")
        (tmp_path / "prefs.js").write_text("some prefs")
        job.config_testcase(str(tmp_path))
        assert os.path.normpath(job.target.prefs) \
            != os.path.normpath(str(tmp_path / "prefs.js"))
        with open(job.target.prefs) as prefs_fp:
            assert prefs_fp.read() == "some prefs"
        assert job.result_code is None


def test_config_testcase_11(tmp_path):
    """env vars from testcase are used"""
    with TestReductionJob(tmp_path, create_binary=False) as job:
        with open(str(tmp_path / "test_info.json"), "w") as info:
            json.dump({
                "target": "test.html",
                "env": {
                    "foo": "bar",
                    "var": "value"
                }}, info)
        (tmp_path / "test.html").write_text("hello")
        job.config_testcase(str(tmp_path))
        assert job._env_mod == dict(var="value", foo="bar")
        assert job.result_code is None


def test_config_testcase_11_legacy(tmp_path):
    """env vars from testcase are used"""
    # TODO: remove this test
    with TestReductionJob(tmp_path, create_binary=False) as job:
        (tmp_path / "test_info.txt").write_text("landing page: test.html")
        (tmp_path / "test.html").write_text("hello")
        (tmp_path / "env_vars.txt").write_text("var=value\nfoo=bar")
        job.config_testcase(str(tmp_path))
        assert job._env_mod == dict(var="value", foo="bar")
        assert job.result_code is None


def test_config_testcase_12(tmp_path):
    """html testcase is loaded ok"""
    with TestReductionJob(tmp_path, create_binary=False) as job:
        (tmp_path / "test.html").write_text("hello")
        job.config_testcase(str(tmp_path / "test.html"))
        assert job._testcase == os.path.join(job._tcroot, "test.html")
        with open(job._testcase) as tc_fp:
            assert tc_fp.read() == "hello"
        assert job.result_code is None


def test_config_testcase_13(tmp_path):
    """test test_info.json not test_info.txt is used"""
    # TODO: remove this test
    with TestReductionJob(tmp_path, create_binary=False) as job:
        (tmp_path / "test_info.txt").write_text("landing page: test_old.html")
        (tmp_path / "test_old.html").write_text("fail!")
        (tmp_path / "env_vars.txt").write_text("var=fail\nfoo=fail")
        with open(str(tmp_path / "test_info.json"), "w") as info:
            json.dump({
                "target": "test.html",
                "env": {
                    "foo": "bar",
                    "var": "value"
                }}, info)
        (tmp_path / "test.html").write_text("hello")
        job.config_testcase(str(tmp_path))
        assert job._env_mod == dict(var="value", foo="bar")
        assert job._testcase == os.path.join(job._tcroot, "test.html")
        with open(job._testcase) as tc_fp:
            assert tc_fp.read() == "hello"
        assert job.result_code is None


def test_run_01(tmp_path):
    """single required testcase is reduced and reported"""
    with TestReductionJob(tmp_path) as job:
        (tmp_path / "tc").mkdir()
        with open(str(tmp_path / "tc" / "test_info.json"), "w") as info:
            json.dump({
                "target": "test.html",
                "env": {
                    "foo": "bar",
                    "var": "value"
                }}, info)
        (tmp_path / "tc" / "test.html").write_bytes(b"fluff\nrequired\n")
        (tmp_path / "tc" / "prefs.js").write_bytes(b"some prefs")
        job.config_testcase(str(tmp_path / "tc"))
        report_data = {"num_reports": 0}

        class FakeReporter(BaseFakeReporter):
            def _submit_report(self, _report, test_cases):
                assert len(test_cases) == 1, "too many test_cases: %r" % (test_cases,)
                tc = test_cases[0]
                assert len(tc._files.required) == 1, \
                    "too many test_files: %r" % (tc._files.required,)
                assert tc.landing_page == "test.html"
                assert tc._files.required[0].data == b"required\n"
                prefs_data = None
                for meta_file in tc._files.meta:
                    if meta_file.file_name == "prefs.js":
                        prefs_data = meta_file.data
                        break
                assert prefs_data == b"some prefs"
                assert tc.env_vars == dict(var="value", foo="bar")
                assert self.quality == FuzzManagerReporter.QUAL_REDUCED_RESULT
                assert self.force_report
                report_data["num_reports"] += 1
        job.set_reporter(FakeReporter())

        assert job.run()
        assert report_data["num_reports"] == 1


def test_run_02(tmp_path):
    """other crashes are reported as unreduced crashes"""
    with TestReductionJobAlt(tmp_path) as job:
        (tmp_path / "tc").mkdir()
        with open(str(tmp_path / "tc" / "test_info.json"), "w") as info:
            json.dump({
                "target": "test.html",
                "env": {
                    "foo": "bar",
                    "var": "value"
                }}, info)
        (tmp_path / "tc" / "test.html").write_bytes(b"fluff\nrequired\n")
        (tmp_path / "tc" / "prefs.js").write_bytes(b"some prefs")
        job.config_testcase(str(tmp_path / "tc"))
        report_data = {"num_reports": 0}

        class FakeReporter(BaseFakeReporter):
            def _submit_report(self, _report, test_cases):
                assert len(test_cases) == 1, "too many test_cases: %r" % (test_cases,)
                tc = test_cases[0]
                assert len(tc._files.required) == 1, \
                    "too many test_files: %r" % (tc._files.required,)
                assert tc._files.required[0].data == b"required\n"
                assert tc.landing_page == "test.html"
                prefs_data = None
                for meta_file in tc._files.meta:
                    if meta_file.file_name == "prefs.js":
                        prefs_data = meta_file.data
                        break
                assert prefs_data == b"some prefs"
                assert tc.env_vars == dict(var="value", foo="bar")
                assert self.quality == FuzzManagerReporter.QUAL_UNREDUCED
                assert not self.force_report
                report_data["num_reports"] += 1
        job.set_reporter(FakeReporter())

        assert not job.run()
        assert report_data["num_reports"] == 1


def test_run_03(tmp_path):
    """other files in testcase are not reduced without DDBEGIN/END"""
    with TestReductionJob(tmp_path) as job:
        (tmp_path / "tc").mkdir()
        (tmp_path / "tc" / "test_info.json").write_text("{\"target\":\"test.html\",\"env\":{}}")
        (tmp_path / "tc" / "test.html").write_bytes(b"fluff\nrequired\n")
        (tmp_path / "tc" / "test2.html").write_bytes(b"fluff\nrequired\n")
        job.config_testcase(str(tmp_path / "tc"))
        report_data = {"num_reports": 0}

        class FakeReporter(BaseFakeReporter):
            def _submit_report(self, _report, test_cases):
                assert len(test_cases) == 1, "too many test_cases: %r" % (test_cases,)
                tc = test_cases[0]
                assert len(tc._files.required) == 1, \
                    "expecting 1 test_file: %r" % ({f.file_name for f in tc._files.required},)
                assert len(tc._files.optional) == 1, \
                    "expecting 1 test_file: %r" % ({f.file_name for f in tc._files.optional},)
                assert tc.landing_page == "test.html"
                assert {x.file_name: x.data for x in tc._files.required} \
                    == {"test.html": b"required\n"}
                assert {x.file_name: x.data for x in tc._files.optional} \
                    == {"test2.html": b"fluff\nrequired\n"}
                assert self.quality == FuzzManagerReporter.QUAL_REDUCED_RESULT
                assert self.force_report
                report_data["num_reports"] += 1
        job.set_reporter(FakeReporter())

        assert job.run()
        assert report_data["num_reports"] == 1


def test_run_04(tmp_path):
    """other files in testcase are reduced with DDBEGIN/END"""
    with TestReductionJob(tmp_path) as job:
        (tmp_path / "tc").mkdir()
        (tmp_path / "tc" / "test_info.json").write_text("{\"target\":\"test.html\",\"env\":{}}")
        (tmp_path / "tc" / "test.html").write_bytes(b"fluff\nrequired\n")
        (tmp_path / "tc" / "test2.html").write_bytes(b"DDBEGIN\nfluff\nrequired\nDDEND\n")
        job.config_testcase(str(tmp_path / "tc"))
        report_data = {"num_reports": 0}

        class FakeReporter(BaseFakeReporter):
            def _submit_report(self, _report, test_cases):
                assert len(test_cases) == 1, "too many test_cases: %r" % (test_cases,)
                tc = test_cases[0]
                assert len(tc._files.required) == 1, \
                    "expecting 1 test file: %r" % (tc._files.required,)
                assert len(tc._files.optional) == 1, \
                    "expecting 1 test file: %r" % (tc._files.optional,)
                assert tc.landing_page == "test.html"
                assert {x.file_name: x.data for x in tc._files.required} \
                    == {"test.html": b"required\n"}
                assert {x.file_name: x.data for x in tc._files.optional} \
                    == {"test2.html": b"DDBEGIN\nrequired\nDDEND\n"}
                assert self.quality == FuzzManagerReporter.QUAL_REDUCED_RESULT
                assert self.force_report
                report_data["num_reports"] += 1
        job.set_reporter(FakeReporter())

        assert job.run()
        assert report_data["num_reports"] == 1


def test_run_05(tmp_path):
    """multiple testcases result in harness being reported"""
    with TestReductionJobKeepHarness(tmp_path) as job:
        (tmp_path / "tc").mkdir()
        (tmp_path / "tc" / "-0").mkdir()
        (tmp_path / "tc" / "-1").mkdir()
        (tmp_path / "tc" / "-0" / "test_info.json").write_text("{\"target\":\"required.html\",\"env\":{}}")
        (tmp_path / "tc" / "-0" / "required.html").write_bytes(b"DDBEGIN\nfluff\nrequired\nDDEND\n")
        (tmp_path / "tc" / "-1" / "test_info.json").write_text("{\"target\":\"required.html\",\"env\":{}}")
        (tmp_path / "tc" / "-1" / "required.html").write_bytes(b"DDBEGIN\nfluff\nrequired\nDDEND\n")
        job.config_testcase(str(tmp_path / "tc"))
        report_data = {"num_reports": 0}

        class FakeReporter(BaseFakeReporter):
            def _submit_report(self, _report, test_cases):
                assert len(test_cases) == 1, "too many test_cases: %r" % (test_cases,)
                tc = test_cases[0]
                assert len(tc._files.required) == 1, \
                    "expecting 1 test file: %r" % (tc._files.required,)
                assert len(tc._files.optional) == 2, \
                    "expecting 2 test files: %r" % (tc._files.optional,)
                assert tc.landing_page.startswith("harness_")
                assert tc.landing_page.endswith(".html")
                assert {x.file_name: x.data for x in tc._files.optional} \
                    == {os.path.join("-0", "required.html"): b"DDBEGIN\nrequired\nDDEND\n",
                        os.path.join("-1", "required.html"): b"DDBEGIN\nrequired\nDDEND\n"}
                assert self.quality == FuzzManagerReporter.QUAL_REDUCED_RESULT
                assert self.force_report
                report_data["num_reports"] += 1
        job.set_reporter(FakeReporter())

        assert job.run()
        assert report_data["num_reports"] == 1


def test_run_06(tmp_path):
    """multiple testcases reducing to 1 file will have harness removed"""
    with TestReductionJob(tmp_path) as job:
        (tmp_path / "tc").mkdir()
        (tmp_path / "tc" / "-0").mkdir()
        (tmp_path / "tc" / "-1").mkdir()
        (tmp_path / "tc" / "-0" / "test_info.json").write_text("{\"target\":\"test.html\",\"env\":{}}")
        (tmp_path / "tc" / "-0" / "test.html").write_bytes(b"-0\nDDBEGIN\nfluff\nrequired\nDDEND\n")
        (tmp_path / "tc" / "-1" / "test_info.json").write_text("{\"target\":\"required.html\",\"env\":{}}")
        (tmp_path / "tc" / "-1" / "required.html").write_bytes(b"-1\nDDBEGIN\nfluff\nrequired\nDDEND\n")
        job.config_testcase(str(tmp_path / "tc"))
        report_data = {"num_reports": 0}

        class FakeReporter(BaseFakeReporter):
            def _submit_report(self, _report, test_cases):
                assert len(test_cases) == 1, "too many test_cases: %r" % (test_cases,)
                tc = test_cases[0]
                assert len(tc._files.required) == 1, \
                    "too many test_files: %r" % (tc._files.required,)
                assert tc.landing_page == "required.html"
                assert tc._files.required[0].data == b"-1\nDDBEGIN\nrequired\nDDEND\n"
                assert self.quality == FuzzManagerReporter.QUAL_REDUCED_RESULT
                assert self.force_report
                report_data["num_reports"] += 1
        job.set_reporter(FakeReporter())

        assert job.run()
        assert report_data["num_reports"] == 1


@pytest.mark.skipif(not strategies.HAVE_JSBEAUTIFIER, reason="jsbeautifier required")
def test_run_07(tmp_path):
    """test that jsbeautifier stage works"""
    with TestReductionJob(tmp_path) as job:
        (tmp_path / "tc").mkdir()
        (tmp_path / "tc" / "test_info.json").write_text("{\"target\":\"test.js\",\"env\":{}}")
        (tmp_path / "tc" / "test.js").write_text("try{'fluff';'required'}catch(e){}\n")
        job.config_testcase(str(tmp_path / "tc"))
        report_data = {"num_reports": 0}

        class FakeReporter(BaseFakeReporter):
            def _submit_report(self, _report, test_cases):
                assert len(test_cases) == 1, "too many test_cases: %r" % (test_cases,)
                tc = test_cases[0]
                assert len(tc._files.required) == 1, \
                    "too many test_files: %r" % (tc._files.required,)
                assert tc.landing_page == "test.js"
                result = tc._files.required[0].data.decode("UTF-8").lstrip(" ")
                assert result == "'required'%s" % (str(os.linesep),)
                assert self.quality == FuzzManagerReporter.QUAL_REDUCED_RESULT
                assert self.force_report
                report_data["num_reports"] += 1
        job.set_reporter(FakeReporter())

        assert job.run()
        assert report_data["num_reports"] == 1


@pytest.mark.skipif(not strategies.HAVE_CSSBEAUTIFIER, reason="cssbeautifier required")
def test_run_08(tmp_path):
    """test that cssbeautifier stage works with .css file"""
    with TestReductionJob(tmp_path) as job:
        (tmp_path / "tc").mkdir()
        (tmp_path / "tc" / "test_info.json").write_text("{\"target\":\"test.css\",\"env\":{}}")
        (tmp_path / "tc" / "test.css").write_text("*,#a{fluff:0;required:1}\n")
        job.config_testcase(str(tmp_path / "tc"))
        report_data = {"num_reports": 0}

        class FakeReporter(BaseFakeReporter):
            def _submit_report(self, _report, test_cases):
                assert len(test_cases) == 1, "too many test_cases: %r" % (test_cases,)
                tc = test_cases[0]
                assert len(tc._files.required) == 1, \
                    "too many test_files: %r" % (tc._files.required,)
                assert tc.landing_page == "test.css"
                result = tc._files.required[0].data.decode("UTF-8").rstrip()
                assert result == "  required: 1"
                assert self.quality == FuzzManagerReporter.QUAL_REDUCED_RESULT
                assert self.force_report
                report_data["num_reports"] += 1
        job.set_reporter(FakeReporter())

        assert job.run()
        assert report_data["num_reports"] == 1

@pytest.mark.skipif(not strategies.HAVE_CSSBEAUTIFIER, reason="cssbeautifier required")
def test_run_09(tmp_path):
    """test that cssbeautifier stage works with .html file"""
    # TODO: tests specifically targeting beautifiers should be created
    with TestReductionJob(tmp_path) as job:
        (tmp_path / "tc").mkdir()
        (tmp_path / "tc" / "test_info.json").write_text("{\"target\":\"test.html\",\"env\":{}}")
        (tmp_path / "tc" / "test.html").write_text("<style>*,#a{fluff:0;required:1}</style>\n")
        job.config_testcase(str(tmp_path / "tc"))
        report_data = {"num_reports": 0}

        class FakeReporter(BaseFakeReporter):
            def _submit_report(self, _report, test_cases):
                assert len(test_cases) == 1, "too many test_cases: %r" % (test_cases,)
                tc = test_cases[0]
                assert len(tc._files.required) == 1, \
                    "too many test_files: %r" % (tc._files.required,)
                assert tc.landing_page == "test.html"
                result = tc._files.required[0].data.decode("UTF-8").rstrip()
                assert result == "  required: 1"
                assert self.quality == FuzzManagerReporter.QUAL_REDUCED_RESULT
                assert self.force_report
                report_data["num_reports"] += 1
        job.set_reporter(FakeReporter())

        assert job.run()
        assert report_data["num_reports"] == 1

def test_run_10(tmp_path):
    """test that jschar stage works"""
    with TestReductionJob(tmp_path) as job:
        (tmp_path / "tc").mkdir()
        (tmp_path / "tc" / "test_info.json").write_bytes(b"{\"target\":\"test.js\",\"env\":{}}")
        (tmp_path / "tc" / "test.js").write_bytes(b"var x = 'xrequiredx'\n")
        job.config_testcase(str(tmp_path / "tc"))
        report_data = {"num_reports": 0}

        class FakeReporter(BaseFakeReporter):
            def _submit_report(self, _report, test_cases):
                assert len(test_cases) == 1, "too many test_cases: %r" % (test_cases,)
                tc = test_cases[0]
                assert len(tc._files.required) == 1, \
                    "too many test_files: %r" % (tc._files.required,)
                assert tc.landing_page == "test.js"
                # strip() is required because jsbeautifier stage removes the newline (if installed)
                result = tc._files.required[0].data.decode("UTF-8").strip()
                assert result == "var x = 'required'"
                assert self.quality == FuzzManagerReporter.QUAL_REDUCED_RESULT
                assert self.force_report
                report_data["num_reports"] += 1
        job.set_reporter(FakeReporter())

        assert job.run()
        assert report_data["num_reports"] == 1


def test_run_11(tmp_path):
    """test that analyze stage works"""
    with TestReductionJobSemiReliable(tmp_path, skip_analysis=False) as job:
        (tmp_path / "tc").mkdir()
        (tmp_path / "tc" / "test_info.json").write_text("{\"target\":\"test.html\",\"env\":{}}")
        (tmp_path / "tc" / "test.html").write_text("fluff\nrequired\n")
        (tmp_path / "tc" / "prefs.js").write_text("some prefs")
        (tmp_path / "tc" / "env_vars.txt").write_text("var=value\nfoo=bar")
        job.config_testcase(str(tmp_path / "tc"))

        job.set_reporter(BaseFakeReporter())

        # try a 50% reliable testcase
        with job.analysis_mode() as params:
            job.test_set_n(5)
            job.run()
            assert params.min_crashes == 2
            assert params.repeat == 10
            assert not params.no_harness

        # try a 90% reliable testcase
        with job.analysis_mode() as params:
            job.test_set_n(9)
            job.run()
            assert params.min_crashes == 2
            assert params.repeat == 4
            assert not params.no_harness

        # try a 100% reliable testcase that doesn't repro with the harness
        with job.analysis_mode() as params:
            job.test_set_n(11, require_no_harness=True)
            job.run()
            assert params.min_crashes == 2
            assert params.repeat == 2
            assert params.no_harness


def test_run_12(tmp_path):
    """test that analyze stage works with multiple testcases"""
    with TestReductionJobSemiReliable(tmp_path, testcase_cache=True, skip_analysis=False) as job:
        (tmp_path / "tc").mkdir()
        (tmp_path / "tc" / "-0").mkdir()
        (tmp_path / "tc" / "-1").mkdir()
        (tmp_path / "tc" / "-0" / "test_info.json").write_text("{\"target\":\"test.html\",\"env\":{}}")
        (tmp_path / "tc" / "-0" / "test.html").write_text("-0\nDDBEGIN\nfluff\nrequired\nDDEND\n")
        (tmp_path / "tc" / "-1" / "test_info.json").write_text("{\"target\":\"required.html\",\"env\":{}}")
        (tmp_path / "tc" / "-1" / "required.html").write_text("-1\nDDBEGIN\nfluff\nrequired\nDDEND\n")
        job.config_testcase(str(tmp_path / "tc"))

        job.set_reporter(BaseFakeReporter())

        with job.analysis_mode() as params:
            params.min_crashes = 1
            params.repeat = 1
            job.test_set_n(5)
            job.run()
            assert params.min_crashes == 2
            assert params.repeat == 10
            assert not params.no_harness



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from collections import deque
from os import environ
from os.path import isfile

from sapphire.server_map import ServerMap
from .storage import TestCase, TestFile
from ..target import sanitizer_opts


__all__ = ("IOManager",)
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]


class IOManager(object):
    TRACKED_ENVVARS = (
        "ASAN_OPTIONS",
        "LSAN_OPTIONS",
        "GNOME_ACCESSIBILITY",
        "GRZ_FORCED_CLOSE",
        "MOZ_CHAOSMODE",
        "XPCOM_DEBUG_BREAK")

    def __init__(self, report_size=1):
        assert report_size > 0
        self.harness = None
        self.server_map = ServerMap()  # manage redirects, include directories and dynamic responses
        self.tests = deque()
        self._environ_files = list()  # collection of files that should be added to the testcase
        self._generated = 0  # number of test cases generated
        self._report_size = report_size
        # used to record environment variable that directly impact the browser
        self._tracked_env = self.tracked_environ()
        self._add_suppressions()

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.cleanup()

    def _add_suppressions(self):
        # Add suppression files to environment files
        for env_var in (x for x in environ if "SAN_OPTIONS" in x):
            opts = sanitizer_opts(environ.get(env_var, ""))
            if "suppressions" not in opts:
                continue
            supp_file = opts["suppressions"].strip("'\"")
            if isfile(supp_file):
                fname = "%s.supp" % (env_var.split("_")[0].lower(),)
                self._environ_files.append(TestFile.from_file(supp_file, fname))

    def cleanup(self):
        for e_file in self._environ_files:
            e_file.close()
        self.purge_tests()

    def create_testcase(self, adapter_name):
        # create testcase object and landing page names
        test = TestCase(
            self.page_name(),
            self.page_name(offset=1),
            adapter_name=adapter_name)
        # add environment variable info to the test case
        for e_name, e_value in self._tracked_env.items():
            test.add_environ_var(e_name, e_value)
        # add environment files to the test case
        for e_file in self._environ_files:
            test.add_meta(e_file.clone())
        # reset redirect map
        self.server_map.redirect.clear()
        self.server_map.set_redirect("grz_current_test", self.page_name(), required=False)
        self.server_map.set_redirect("grz_next_test", self.page_name(offset=1))
        if self.harness is not None:
            # add harness to testcase
            self.server_map.set_dynamic_response(
                "grz_harness",
                lambda: self.harness,
                mime_type="text/html")
        self._generated += 1
        self.tests.append(test)
        # manage testcase cache size
        if len(self.tests) > self._report_size:
            self.tests.popleft().cleanup()
        return test

    def page_name(self, offset=0):
        return "test_%04d.html" % (self._generated + offset,)

    def purge_tests(self):
        for testcase in self.tests:
            testcase.cleanup()
        self.tests.clear()

    @staticmethod
    def tracked_environ():
        # Scan os.environ and collect environment variables
        # that are relevant to Grizzly or the test case.
        env = dict()
        tracked_san_opts = ("detect_leaks",)
        for var in IOManager.TRACKED_ENVVARS:
            if var not in environ:
                continue
            if var.endswith("SAN_OPTIONS"):
                opts = sanitizer_opts(environ.get(var, ""))
                # strip unwanted options
                tracked = dict()
                for opt in tracked_san_opts:
                    if opt in opts:
                        tracked[opt] = opts[opt]
                # only record *SAN_OPTIONS if there are options
                if tracked:
                    env[var] = ":".join("=".join((k, v)) for k, v in tracked.items())
            elif var == "XPCOM_DEBUG_BREAK" and environ.get(var, "").lower() == "warn":
                # ignore FFPuppet default XPCOM_DEBUG_BREAK value (set in helpers.py)
                continue
            else:
                env[var] = environ[var]
        return env



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import pytest

from .stack_hasher import Stack, StackFrame


def test_stack_01():
    """test creating an empty Stack"""
    stack = Stack()
    assert stack.minor is None
    assert isinstance(stack.frames, list)
    assert stack._major_depth > 0  # pylint: disable=protected-access

def test_stack_02():
    """test creating a Stack with 1 frame"""
    frames = [StackFrame(function="a", location="b", offset="c", stack_line="0")]
    stack = Stack(frames=frames)
    assert stack.minor is not None
    assert stack.major is not None
    # at this point the hashes should match
    assert stack.minor == stack.major
    assert len(stack.frames) == 1
    output = str(stack)
    assert "00" in output
    assert "function: 'a'" in output
    assert "location: 'b'" in output
    assert "offset: 'c'" in output

def test_stack_03():
    """test creating a Stack with 2 frames"""
    frames = list()
    for _ in range(2):
        frames.append(StackFrame(function="a", location="b", offset="c", stack_line="0"))
    stack = Stack(frames=frames, major_depth=2)
    assert stack.minor is not None
    assert stack.major is not None
    # at this point the hashes should not match because offset on the major hash is
    # only added from the top frame
    assert stack.minor != stack.major
    assert len(stack.frames) == 2

def test_stack_04():
    """test creating a Stack with 2 frames with a major depth of 0"""
    frames = list()
    for line in range(2):
        frames.append(StackFrame(function="a", location="b", offset="c", stack_line=str(line)))
    stack = Stack(frames=frames, major_depth=0)
    assert stack.minor is not None
    assert stack.major is None
    # at this point the hashes should not match because offset on the major hash is
    # only added from the top frame
    assert stack.minor != stack.major
    assert len(stack.frames) == 2

def test_stack_05():
    """test creating a Stack with 10 frames exceeding major depth"""
    frames = list()
    for line in range(10):
        frames.append(StackFrame(function="a", location="b", offset="c", stack_line=str(line)))
    stack = Stack(frames=frames, major_depth=5)
    assert stack.minor is not None
    assert stack.major is not None
    # at this point the hashes should not match because offset on the major hash is
    # only added from the top frame
    assert stack.minor != stack.major
    assert len(stack.frames) == 10
    assert stack.major != Stack(frames=frames, major_depth=4).major

def test_stack_06():
    """test creating a Stack by calling from_text()"""
    input_txt = "" \
        "=================================================================\n" \
        "==7854==ERROR: AddressSanitizer: SEGV on unknown address 0x000000000000 (pc 0x7fcca620e0ee bp 0x7ffd946b2690 sp 0x7ffd946b25c0 T0)\n" \
        "==7854==blah.\n" \
        "==7854==Hint: blah.\n" \
        "    #0 0x7fcca620e0ed in test::test::test(nsIWa*, nsICa*) /aa/bb/cc/dd/ee/ff/gg/asdf.cpp:5533:14\n" \
        "    #1 0x7fcca3497201 in nsAs::TestTest(mz::dom::El*, int) /aa/bb/cc/dd/ee/ff/nsFilea.cpp:13733:3\n" \
        "    #2 0x7fcca3495b9b in asdf::fasd() /aa/bb/cc/dd/ee/ff/base/nsFileb.cpp:11674:21\n" \
        "    #3 0x7fcca02eb9d8 in nsAasd::PrNeEv(bool, bool*) /aa/bb/cc/dd/ee/xpcom/nsFilec.cpp:1396:14\n" \
        "    #4 0x7fcca0307d40 in R_PNE(asd*, bool) /aa/bb/cc/dd/ee/xpcom/threads/asf.cpp:657:10\n" \
        "    #5 0x2a780b25f65a  (<unknown module>)\n" \
        "\n" \
        "AddressSanitizer can not provide additional info.\n" \
        "SUMMARY: AddressSanitizer: SEGV /aa/bb/cc/dd/ee/ff/asdf.cpp:5533:14 in test::test::test(nsIWa*, nsICa*)\n" \
        "==7854==ABORTING\n"
    stack = Stack.from_text(input_txt)
    assert len(stack.frames) == 6
    assert stack.minor != stack.major
    assert stack.frames[0].mode == StackFrame.MODE_SANITIZER

def test_stack_07():
    """test creating a Stack by calling from_text() with mixed frames modes"""
    input_txt = "" \
    "    #0 0x4d2cde in a_b_c /a/lib/info.c:392:12\n" \
    "    #1 0x491e82 in main /a/b/d_e.c:128:8\n" \
    "    #2 0x7f090384582f in __libc_start_main /build/glibc-glibc-2.23/csu/../csu/libc-start.c:291\n" \
    "#2  0x0000000000400545 in gdb_frame ()\n" \
    "    #3 0x41b228 in _start (bin_name+0x41b228)\n"
    stack = Stack.from_text(input_txt)
    assert len(stack.frames) == 4
    assert stack.minor != stack.major
    assert stack.frames[0].mode == StackFrame.MODE_SANITIZER

def test_stack_08():
    """test creating a Stack by calling from_text() with text containing 2 stacks"""
    input_txt = "" \
        "    #0 0x0bad0bad in bad::frame0(nsA*, nsB*) /aa/a.cpp:12:1\n" \
        "    #1 0x0bad0bad in bad::frame1(mz::d::EE*, int) /aa/a.cpp:12:1\n" \
        "    #0 0x1badf00d in good::frame0(nsA*, nsB*) /aa/a.cpp:12:1\n" \
        "    #1 0xdeadbeef in good::frame1(mz::d::EE*, int) /aa/a.cpp:12:1\n"
    stack = Stack.from_text(input_txt)
    assert len(stack.frames) == 2
    assert stack.frames[0].function == "good::frame0"
    assert stack.frames[1].function == "good::frame1"
    assert stack.minor != stack.major
    assert stack.frames[0].mode == StackFrame.MODE_SANITIZER

def test_stack_09():
    """test creating a Stack by calling from_text() with empty string"""
    stack = Stack.from_text("")
    assert not stack.frames
    assert stack.minor is None
    assert stack.major is None

def test_stack_10():
    """test creating a Stack from a Sanitizer trace with an unsymbolized lib"""
    input_txt = "" \
        "    #0 0x4c7702 in realloc asan/asan_malloc_linux.cc:107:3\n" \
        "    #1 0x7f6d056ce7fc  (/lib/x86_64-linux-gnu/libdbus-1.so.3+0x2d7fc)\n" \
        "    #2 0x7ffffffff  (<unknown module>)\n"
    stack = Stack.from_text(input_txt)
    assert len(stack.frames) == 3
    assert stack.frames[0].location == "asan_malloc_linux.cc"
    assert stack.frames[1].location == "libdbus-1.so.3"
    assert stack.frames[2].location == "<unknown module>"
    assert stack.minor != stack.major
    assert stack.frames[0].mode == StackFrame.MODE_SANITIZER

def test_stack_11():
    """test creating a Stack from a Sanitizer trace with an unsymbolized lib"""
    input_txt = "" \
        "    #0 0x90000223  (/usr/swr_a.so+0x231223)\n" \
        "    #1 0x00000447  (/usr/as.so.1+0x42447)\n" \
        "    #2 0x000098fc in fSasd /src/obj-firefox/dist/include/something.h:102:9\n" \
        "    #3 0x000098fc in mz::as::asdf::SB() /src/Blah.cpp:655\n"
    stack = Stack.from_text(input_txt)
    assert len(stack.frames) == 4
    assert stack.frames[0].location == "swr_a.so"
    assert stack.frames[1].location == "as.so.1"
    assert stack.frames[2].function == "fSasd"
    assert stack.frames[3].function == "mz::as::asdf::SB"
    assert stack.minor != stack.major
    assert stack.frames[0].mode == StackFrame.MODE_SANITIZER

def test_stack_12():
    """test creating a Stack from a Valgrind trace"""
    input_txt = "" \
        "==4754== \n" \
        "==4754== Use of uninitialised value of size 8\n" \
        "==4754==    at 0x45C6C0: FooBar (decode.c:964)\n" \
        "==4754==    by 0x462A20: main (test.cc:71)\n" \
        "==4754==  Uninitialised value was created by a heap allocation\n" \
        "==4754==    at 0x4C2AB80: malloc (in /usr/lib/test-linux.so)\n" \
        "==4754==    by 0x459455: FooBar (decode.c:757)\n" \
        "==4754==    by 0x462A20: main (test.cc:71)\n" \
        "==4754== \n"
    stack = Stack.from_text(input_txt)
    assert len(stack.frames) == 5
    assert stack.frames[0].location == "decode.c"
    assert stack.frames[1].location == "test.cc"
    assert stack.frames[2].function == "malloc"
    assert stack.frames[3].function == "FooBar"
    assert stack.frames[4].function == "main"
    assert stack.minor != stack.major
    assert stack.frames[0].mode == StackFrame.MODE_VALGRIND

def test_stack_13():
    """test creating a Stack from Rust trace"""
    input_txt = "" \
    "thread '<unnamed>' panicked at 'Invoking Servo_Element_IsDisplayContents on unstyled element', libcore/option.rs:917:5\n" \
    "stack backtrace:\n" \
    "   0:     0x7ff1c65e93d3 - std::sys::unix::backtrace::tracing::imp::unwind_backtrace::h09c1ee131a74b1c4\n" \
    "                               at libstd/sys/unix/backtrace/tracing/gcc_s.rs:49\n" \
    "   1:     0x7ff1c65e81c4 - std::panicking::default_hook::{{closure}}::h945a649c9017832e\n" \
    "                               at libstd/sys_common/backtrace.rs:71\n" \
    "                               at libstd/sys_common/backtrace.rs:59\n" \
    "                               at libstd/panicking.rs:380\n" \
    "   2:     0x7ff1c65e7457 - std::panicking::default_hook::hcc534c2d30fbcda3\n" \
    "                               at libstd/panicking.rs:396\n" \
    "   3:     0x7ff1c65e6de7 - std::panicking::rust_panic_with_hook::h09a7a3a353dc2f38\n" \
    "                               at libstd/panicking.rs:576\n" \
    "   4:     0x7ff1c65e6c95 - std::panicking::begin_panic::h8327f16bde15df70\n" \
    "                               at libstd/panicking.rs:537\n" \
    "   5:     0x7ff1c65e6c29 - std::panicking::begin_panic_fmt::h42ff1d37404632d6\n" \
    "                               at libstd/panicking.rs:521\n" \
    "   6:     0x7ff1c65fa46a - core::panicking::panic_fmt::h0bd854df201d1baf\n" \
    "                               at libstd/panicking.rs:497\n" \
    "   7:     0x7ff1c65ffba8 - core::option::expect_failed::hfa0c8a51e07f7adc\n" \
    "                               at libcore/option.rs:917\n" \
    "   8:     0x7ff1c632d473 - Servo_Element_IsDisplayContents\n" \
    "                               at /checkout/src/libcore/option.rs:302\n" \
    "                               at servo/ports/geckolib/glue.rs:1086\n" \
    "   9:     0x7f44064ee749 - _ZNK7nsFrame24DoGetParentComputedStyleEPP8nsIFrame\n" \
    "                                at /builds/worker/workspace/build/src/layout/generic/nsFrame.cpp:9893\n" \
    "   10:     0x7f4406229749 - _ZN7mozilla14RestyleManager35DoReparentComputedStyleForFirstLineEP8nsIFrameRNS_13ServoStyleSetE\n" \
    "                                at /builds/worker/workspace/build/src/layout/base/RestyleManager.cpp:3407\n" \
    "   11:     0x7f440622a0a5 - _ZN7mozilla14RestyleManager24ReparentFrameDescendantsEP8nsIFrameS2_RNS_13ServoStyleSetE\n" \
    "                                at /builds/worker/workspace/build/src/layout/base/RestyleManager.cpp:3538\n"
    stack = Stack.from_text(input_txt)
    assert len(stack.frames) == 12
    assert stack.frames[0].function == "std::sys::unix::backtrace::tracing::imp::unwind_backtrace"
    assert stack.frames[8].function == "Servo_Element_IsDisplayContents"
    assert stack.minor != stack.major
    assert stack.frames[0].mode == StackFrame.MODE_RUST

def test_stackframe_01():
    """test creating an empty StackFrame"""
    stack = StackFrame()
    assert not stack.__str__()

def test_stackframe_02():
    """test creating a StackFrame from junk"""
    assert StackFrame.from_line("#0      ") is None
    assert StackFrame.from_line(" #0 ") is None
    with pytest.raises(AssertionError) as exc:
        StackFrame.from_line("#0 \n \n\n\n#1\n\ntest()!")
    assert "Input contains unexpected new line(s)" in str(exc.value)
    assert StackFrame.from_line("#0#0#0#0#0#0#0#0") is None
    assert StackFrame.from_line("#a") is None
    assert StackFrame.from_line("") is None
    assert StackFrame.from_line("###") is None
    assert StackFrame.from_line("123") is None
    assert StackFrame.from_line("test()") is None
    assert StackFrame.from_line("|||") is None
    assert StackFrame.from_line("==123==") is None
    assert StackFrame.from_line("==1== by 0x0: a ()") is None

def test_sanitizer_stackframe_01():
    """test creating a StackFrame from a line with symbols"""
    frame = StackFrame.from_line("    #1 0x7f00dad60565 in Abort(char const*) /blah/base/nsDebugImpl.cpp:472")
    assert frame.stack_line == "1"
    assert frame.function == "Abort"
    assert frame.location == "nsDebugImpl.cpp"
    assert frame.offset == "472"
    assert frame.mode == StackFrame.MODE_SANITIZER

def test_sanitizer_stackframe_02():
    """test creating a StackFrame from a line with symbols"""
    frame = StackFrame.from_line("    #36 0x48a6e4 in main /app/nsBrowserApp.cpp:399:11")
    assert frame.stack_line == "36"
    assert frame.function == "main"
    assert frame.location == "nsBrowserApp.cpp"
    assert frame.offset == "399"
    assert frame.mode == StackFrame.MODE_SANITIZER

def test_sanitizer_stackframe_03():
    """test creating a StackFrame from a line without symbols"""
    frame = StackFrame.from_line("    #1 0x7f00ecc1b33f (/lib/x86_64-linux-gnu/libpthread.so.0+0x1033f)")
    assert frame.stack_line == "1"
    assert frame.function is None
    assert frame.location == "libpthread.so.0"
    assert frame.offset == "0x1033f"
    assert frame.mode == StackFrame.MODE_SANITIZER

def test_sanitizer_stackframe_04():
    """test creating a StackFrame from a line with symbols"""
    frame = StackFrame.from_line("    #25 0x7f0155526181 in start_thread (/l/libpthread.so.0+0x8181)")
    assert frame.stack_line == "25"
    assert frame.function == "start_thread"
    assert frame.location == "libpthread.so.0"
    assert frame.offset == "0x8181"
    assert frame.mode == StackFrame.MODE_SANITIZER

def test_sanitizer_stackframe_05():
    """test creating a StackFrame from a line with angle brackets"""
    frame = StackFrame.from_line("    #123 0x7f30afea9148 in Call<nsBlah *> /a/b.cpp:356:50")
    assert frame.stack_line == "123"
    assert frame.function == "Call"
    assert frame.location == "b.cpp"
    assert frame.offset == "356"
    assert frame.mode == StackFrame.MODE_SANITIZER

def test_sanitizer_stackframe_06():
    """test creating a StackFrame from a useless frame"""
    frame = StackFrame.from_line("    #2 0x7ffffffff  (<unknown module>)")
    assert frame.stack_line == "2"
    assert frame.function is None
    assert frame.location == "<unknown module>"
    assert frame.offset is None
    assert frame.mode == StackFrame.MODE_SANITIZER

def test_sanitizer_stackframe_07():
    """test creating a StackFrame from a line missing a function"""
    frame = StackFrame.from_line("    #0 0x7f0d571e04bd  /a/glibc-2.23/../syscall-template.S:84")
    assert frame.stack_line == "0"
    assert frame.function is None
    assert frame.location == "syscall-template.S"
    assert frame.offset == "84"
    assert frame.mode == StackFrame.MODE_SANITIZER

def test_sanitizer_stackframe_08():
    """test creating a StackFrame from a line with lots of spaces"""
    frame = StackFrame.from_line("    #0 0x48a6e4 in Call<a *> /test path/file name.c:1:2")
    assert frame.stack_line == "0"
    assert frame.function == "Call"
    assert frame.location == "file name.c"
    assert frame.offset == "1"
    assert frame.mode == StackFrame.MODE_SANITIZER

def test_gdb_stackframe_01():
    """test creating a StackFrame from a GDB line with symbols"""
    frame = StackFrame.from_line("#0  __memmove_ssse3_back () at ../d/x86_64/a/memcpy-ssse3-back.S:1654")
    assert frame.stack_line == "0"
    assert frame.function == "__memmove_ssse3_back"
    assert frame.location == "memcpy-ssse3-back.S"
    assert frame.offset == "1654"
    assert frame.mode == StackFrame.MODE_GDB

def test_gdb_stackframe_02():
    """test creating a StackFrame from a GDB line with symbols but no line numbers"""
    frame = StackFrame.from_line("#2  0x0000000000400545 in main ()")
    assert frame.stack_line == "2"
    assert frame.function == "main"
    assert frame.location is None
    assert frame.offset is None
    assert frame.mode == StackFrame.MODE_GDB

def test_gdb_stackframe_03():
    """test creating a StackFrame from a GDB line with symbols"""
    frame = StackFrame.from_line("#3  0x0000000000400545 in main () at test.c:5")
    assert frame.stack_line == "3"
    assert frame.function == "main"
    assert frame.location == "test.c"
    assert frame.offset == "5"
    assert frame.mode == StackFrame.MODE_GDB

def test_minidump_stackframe_01():
    """test creating a StackFrame from a Minidump line with symbols"""
    frame = StackFrame.from_line("0|2|libtest|main|hg:c.a.org/m-c:a/b/file.cpp:5bf50|114|0x3a")
    assert frame.stack_line == "2"
    assert frame.function == "main"
    assert frame.location == "file.cpp"
    assert frame.offset == "114"
    assert frame.mode == StackFrame.MODE_MINIDUMP

def test_minidump_stackframe_02():
    """test creating a StackFrame from a Minidump line without symbols"""
    frame = StackFrame.from_line("9|42|libpthread-2.26.so||||0x10588")
    assert frame.stack_line == "42"
    assert frame.function is None
    assert frame.location == "libpthread-2.26.so"
    assert frame.offset == "0x10588"
    assert frame.mode == StackFrame.MODE_MINIDUMP

def test_minidump_stackframe_03():
    """test creating a StackFrame from a Minidump line without hg repo info"""
    frame = StackFrame.from_line("0|49|libxul.so|foo|/usr/x86_64-linux-gnu/test.h|85|0x5")
    assert frame.stack_line == "49"
    assert frame.function == "foo"
    assert frame.location == "/usr/x86_64-linux-gnu/test.h"
    assert frame.offset == "85"
    assert frame.mode == StackFrame.MODE_MINIDUMP

def test_tsan_stackframe_01():
    """test creating a StackFrame from a symbolized TSan line"""
    frame = StackFrame.from_line("    #0 main race.c:10 (exe+0xa3b4)")
    assert frame.stack_line == "0"
    assert frame.function == "main"
    assert frame.location == "race.c"
    assert frame.offset == "10"
    assert frame.mode == StackFrame.MODE_TSAN

def test_tsan_stackframe_02():
    """test creating a StackFrame from a symbolized TSan line"""
    frame = StackFrame.from_line("    #1 test1 test2 /a b/c.h:51:10 (libxul.so+0x18c9873)")
    assert frame.stack_line == "1"
    assert frame.function == "test1"
    assert frame.location == "c.h"
    assert frame.offset == "51"
    assert frame.mode == StackFrame.MODE_TSAN

def test_tsan_stackframe_03():
    """test creating a StackFrame from an unsymbolized TSan line"""
    frame = StackFrame.from_line("    #2 <null> <null> (0xbad)")
    assert frame.stack_line == "2"
    assert frame.function is None
    assert frame.location is None
    assert frame.offset == "0xbad"
    assert frame.mode == StackFrame.MODE_TSAN

def test_tsan_stackframe_04():
    """test creating a StackFrame from a TSan line missing file"""
    frame = StackFrame.from_line("    #0 func <null> (mod+0x123ac)")
    assert frame.stack_line == "0"
    assert frame.function == "func"
    assert frame.location == "mod"
    assert frame.offset == "0x123ac"
    assert frame.mode == StackFrame.MODE_TSAN

def test_valgrind_stackframe_01():
    frame = StackFrame.from_line("==4754==    at 0x45C6C0: FuncName (decode.c:123)")
    assert frame.stack_line is None
    assert frame.function == "FuncName"
    assert frame.location == "decode.c"
    assert frame.offset == "123"
    assert frame.mode == StackFrame.MODE_VALGRIND

def test_valgrind_stackframe_02():
    frame = StackFrame.from_line("==4754==    by 0x462A20: main (foo.cc:71)")
    assert frame.stack_line is None
    assert frame.function == "main"
    assert frame.location == "foo.cc"
    assert frame.offset == "71"
    assert frame.mode == StackFrame.MODE_VALGRIND

def test_valgrind_stackframe_03():
    frame = StackFrame.from_line("==4754==    at 0x4C2AB80: malloc (in /usr/lib/blah-linux.so)")
    assert frame.stack_line is None
    assert frame.function == "malloc"
    assert frame.location == "blah-linux.so"
    assert frame.offset is None
    assert frame.mode == StackFrame.MODE_VALGRIND

def test_valgrind_stackframe_04():
    frame = StackFrame.from_line("==2342==    by 0x4E3E71: (anon ns)::test(b2::a&, int) (main.cpp:49)")
    assert frame.stack_line is None
    assert frame.function == "(anon ns)::test(b2::a&, int)"
    assert frame.location == "main.cpp"
    assert frame.offset == "49"
    assert frame.mode == StackFrame.MODE_VALGRIND

def test_valgrind_stackframe_05():
    frame = StackFrame.from_line("==2342==    at 0xF00D: Foo::Foo(char *, int, bool) (File.h:37)")
    assert frame.stack_line is None
    assert frame.function == "Foo::Foo(char *, int, bool)"
    assert frame.location == "File.h"
    assert frame.offset == "37"
    assert frame.mode == StackFrame.MODE_VALGRIND

def test_rr_stackframe_01():
    frame = StackFrame.from_line("rr(main+0x244)[0x450b74]")
    assert frame.stack_line is None
    assert frame.function is None
    assert frame.location == "main"
    assert frame.offset == "0x244"
    assert frame.mode == StackFrame.MODE_RR

def test_rust_stackframe_01():
    """test creating a Rust StackFrame from stack line"""
    frame = StackFrame.from_line("  53:    0x7ff1d7e4982f - __libc_start_main")
    assert frame.stack_line == "53"
    assert frame.function == "__libc_start_main"
    assert frame.location is None
    assert frame.offset is None
    assert frame.mode == StackFrame.MODE_RUST

def test_rust_stackframe_02():
    """test creating a Rust StackFrame from stack line"""
    frame = StackFrame.from_line("  4:    0x10b715a5b - unwind::begin_unwind_fmt::h227376fe1e021a36n3d")
    assert frame.stack_line == "4"
    assert frame.location is None
    assert frame.function == "unwind::begin_unwind_fmt"
    assert frame.offset is None
    assert frame.mode == StackFrame.MODE_RUST



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from collections import namedtuple
from itertools import chain
import json
import os
import shutil
from tempfile import SpooledTemporaryFile

from ..target import sanitizer_opts
from .utils import grz_tmp


__all__ = ("TestCase", "TestFile", "TestCaseLoadFailure", "TestFileExists")
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]


class TestCaseLoadFailure(Exception):
    """Raised when loading a TestCase fails"""


class TestFileExists(Exception):
    """Raised when adding a TestFile to a TestCase that has an existing TestFile with the same name"""


TestFileMap = namedtuple("TestFileMap", "meta optional required")


class TestCase(object):
    __slots__ = (
        "adapter_name", "duration", "env_vars", "input_fname", "landing_page",
        "redirect_page", "_existing_paths", "_files")

    def __init__(self, landing_page, redirect_page, adapter_name, input_fname=None):
        self.adapter_name = adapter_name
        self.duration = None
        self.env_vars = dict()  # environment variables
        self.input_fname = input_fname  # file that was used to create the test case
        self.landing_page = landing_page
        self.redirect_page = redirect_page
        self._existing_paths = list()  # file paths in use
        self._files = TestFileMap(
            meta=list(),  # environment files such as prefs.js, etc...
            optional=list(),
            required=list())

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.cleanup()

    def _add(self, target, test_file):
        """Add a test file to test case and perform sanity checks.

        Args:
            target (list): Specific list of Files to append target test_file to.
            test_file (TestFile): TestFile to append

        Returns:
            None
        """
        assert isinstance(test_file, TestFile), "only accepts TestFiles"
        if test_file.file_name in self._existing_paths:
            raise TestFileExists("%r exists in test" % (test_file.file_name,))
        self._existing_paths.append(test_file.file_name)
        target.append(test_file)

    def add_batch(self, path, include_files, prefix=None):
        """Iterate over files in include_files and attach the files that are
        located in path to testcase.

        Args:
            path (str): Path to the root of the directory that contains files.
            include_files (iterable): Paths of the files to be added to the
                                      test case if they exist in path.
            prefix (str): Path prefix to prepend to file when adding to
                          test case.

        Returns:
            None
        """
        path = os.path.abspath(path)
        for fname in (x for x in include_files if x.startswith(path)):
            test_path = os.path.relpath(fname, path)
            if test_path.startswith(".."):
                continue
            if prefix:
                test_path = "/".join((prefix, test_path))
            self.add_from_file(fname, file_name=test_path)

    def add_meta(self, meta_file):
        """Add a test file to test case as a meta file.

        Args:
            meta_file (TestFile): TestFile to add to TestCase

        Returns:
            None
        """
        self._add(self._files.meta, meta_file)

    def add_environ_var(self, name, value):
        """Add environment variable to test case.

        Args:
            name (str): Environment variable name
            value (str): Environment variable value

        Returns:
            None
        """
        self.env_vars[name] = value

    def add_file(self, test_file, required=True):
        """Add a test file to test case.

        Args:
            meta_file (TestFile): TestFile to add to TestCase
            required (bool): Indicates if test file must be served

        Returns:
            None
        """
        if required:
            self._add(self._files.required, test_file)
        else:
            self._add(self._files.optional, test_file)

    def add_from_data(self, data, file_name, encoding="UTF-8", required=True):
        """Create a TestFile and add it to the test case.

        Args:
            data (bytes): Data to write to file
            file_name (str): Name for the test file
            encoding (str): Encoding to be used
            required (bool): Indicates if test file must be served

        Returns:
            None
        """
        tfile = TestFile.from_data(data, file_name, encoding=encoding)
        try:
            self.add_file(tfile, required=required)
        except TestFileExists:
            tfile.close()
            raise

    def add_from_file(self, input_file, file_name=None, required=True):
        """Create a TestFile from an existing file and add it to the test case.

        Args:
            input_file (str): Path to existing file to use
            file_name (str): Name for the test file
            required (bool): Indicates if test file must be served

        Returns:
            None
        """
        tfile = TestFile.from_file(input_file, file_name=file_name)
        try:
            self.add_file(tfile, required=required)
        except TestFileExists:
            tfile.close()
            raise

    def cleanup(self):
        """Close all the test files.

        Args:
            None

        Returns:
            None
        """
        for file_group in self._files:
            for test_file in file_group:
                test_file.close()

    def contains(self, file_name):
        """Check TestCase contains the TestFile with name matching `file_name`.

        Args:
            file_name (str): File name to search for in TestCase.

        Returns:
            bool: True if file exists in the TestCase otherwise False
        """
        return file_name in self._existing_paths

    @property
    def data_size(self):
        """The total amount of data used by the test case (bytes).

        Args:
            None

        Returns:
            int: Total size of the test case in byte.
        """
        total = 0
        for group in self._files:
            total += sum(x.size for x in group)
        return total

    def dump(self, out_path, include_details=False):
        """Write all the test case data to the filesystem.

        Args:
            out_path (str): Path to directory to output data
            include_details (bool): Output "test_info.json" file

        Returns:
            None
        """
        # save test files to out_path
        for test_file in chain(self._files.required, self._files.optional):
            test_file.dump(out_path)
        # save test case files and meta data including:
        # adapter used, input file, environment info and files
        if include_details:
            assert isinstance(self.env_vars, dict)
            info = {
                "adapter": self.adapter_name,
                "duration": self.duration,
                "env": self.env_vars,
                "input": os.path.basename(self.input_fname) if self.input_fname else None,
                "target": self.landing_page}
            with open(os.path.join(out_path, "test_info.json"), "w") as out_fp:
                json.dump(info, out_fp, indent=2, sort_keys=True)
            # save meta files
            for meta_file in self._files.meta:
                meta_file.dump(out_path)

    def load_environ(self, path, env_data):
        # sanity check environment variable data
        for name, value in env_data.items():
            if not isinstance(name, str) or not isinstance(value, str):
                raise TestCaseLoadFailure("'env_data' contains invalid 'env' entries")
        self.env_vars = env_data
        known_suppressions = ("lsan.supp", "tsan.supp", "ubsan.supp")
        for supp in os.listdir(path):
            if supp.lower() in known_suppressions:
                # Update *SAN_OPTIONS environment variable to use provided suppression files.
                opt_key = "%s_OPTIONS" % (supp.split(".")[0].upper(),)
                opts = sanitizer_opts(self.env_vars.get(opt_key, ""))
                opts["suppressions"] = "'%s'" % (os.path.join(path, supp),)
                self.env_vars[opt_key] = ":".join("=".join((k, v)) for k, v in opts.items())

    @classmethod
    def load_path(cls, path, full_scan=False, prefs=True):
        """Load contents of a TestCase from disk. If `path` is a directory it must
        contain a valid test_info.json file.

        Args:
            path (str): Path to the directory or file to load.
            full_scan (bool): Include all files in the directory containing the
                              test case entry point as well as the contents of
                              subdirectories. This is always the case when
                              loading a directory.
                              WARNING: This should be used with caution!
            prefs (bool): Include prefs.js file in the test case.

        Returns:
            TestCase: A TestCase.
        """
        path = os.path.abspath(path)
        if os.path.isdir(path):
            # load a directory using test_info.json
            try:
                with open(os.path.join(path, "test_info.json"), "r") as in_fp:
                    info = json.load(in_fp)
            except IOError:
                raise TestCaseLoadFailure("Missing 'test_info.json'")
            except ValueError:
                raise TestCaseLoadFailure("Invalid 'test_info.json'")
            if "target" not in info:
                raise TestCaseLoadFailure("'test_info.json' missing 'target' entry")
            entry_point = os.path.basename(info["target"])
            if not os.path.isfile(os.path.join(path, entry_point)):
                raise TestCaseLoadFailure("entry_point '%s' not found in '%s'" % (entry_point, path))
            adapter = info.get("adapter", None)
            full_scan = True
        elif os.path.isfile(path):
            adapter = None
            entry_point = os.path.basename(path)
            path = os.path.dirname(path)
            info = None
        else:
            raise TestCaseLoadFailure("Cannot find %r" % (path,))
        test = cls(None, None, adapter)
        if full_scan:
            # load all files from directory as test
            for dpath, _, files in os.walk(path):
                for fname in files:
                    if fname == "test_info.json":
                        continue
                    if dpath == path:
                        if fname == "prefs.js":
                            if prefs:
                                test.add_meta(TestFile.from_file(os.path.join(dpath, fname)))
                            continue
                        if fname == entry_point:
                            test.add_from_file(os.path.join(dpath, fname))
                            # set entry point
                            test.landing_page = fname
                            continue
                        location = None
                    else:
                        # handle nested directories
                        location = "/".join((dpath.split(path, 1)[-1], fname))
                    test.add_from_file(
                        os.path.join(dpath, fname),
                        file_name=location,
                        required=False)
        else:
            # load single file as test
            test.add_from_file(os.path.join(path, entry_point))
            test.landing_page = entry_point
        if test.landing_page is None:  # pragma: no cover
            # this should not be possible
            test.cleanup()
            raise AssertionError("Scanning for test case 'entry point' failed")
        # load environment variables
        if info:
            try:
                test.load_environ(path, info.get("env", {}))
            except TestCaseLoadFailure:
                test.cleanup()
                raise
        return test

    @property
    def optional(self):
        """Get file names of optional TestFiles

        Args:
            None

        Returns:
            generator: file names (str) of optional files
        """
        for test in self._files.optional:
            yield test.file_name

    def purge_optional(self, keep):
        """Remove optional files (by name) that are not in keep.

        Args:
            keep (iterable): Filenames that will not be removed.

        Returns:
            None
        """
        keep = set(keep)
        to_remove = []
        for idx, tfile in enumerate(self._files.optional):
            if tfile.file_name not in keep:
                to_remove.append(idx)
        for idx in reversed(to_remove):
            self._files.optional.pop(idx).close()


class TestFile(object):
    CACHE_LIMIT = 0x80000  # data cache limit per file: 512KB
    XFER_BUF = 0x10000  # transfer buffer size: 64KB

    __slots__ = ("_fp", "file_name")

    def __init__(self, file_name):
        if not file_name:
            raise TypeError("TestFile requires a name")
        self._fp = SpooledTemporaryFile(max_size=self.CACHE_LIMIT, dir=grz_tmp(), prefix="grz_tf_")
        # TODO: Add file_name sanitation since it is used when the file is written to the fs
        # XXX: This is a naive fix for a larger path issue
        if "\\" in file_name:
            file_name = file_name.replace("\\", "/")
        if file_name.startswith("/"):
            file_name = file_name.lstrip("/")
        self.file_name = os.path.normpath(file_name)  # name including path relative to wwwroot

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.close()

    def clone(self):
        """Make a copy of the TestFile.

        Args:
            None

        Returns:
            TestFile: A copy of the TestFile instance
        """
        cloned = TestFile(self.file_name)
        self._fp.seek(0)
        shutil.copyfileobj(self._fp, cloned._fp, self.XFER_BUF)  # pylint: disable=protected-access
        return cloned

    def close(self):
        """Close the TestFile.

        Args:
            None

        Returns:
            None TestFile instance
        """
        self._fp.close()

    @property
    def data(self):
        """Get the data from the TestFile. Not recommenced for large files.

        Args:
            None

        Returns:
            bytes: Data from the TestFile
        """
        pos = self._fp.tell()
        self._fp.seek(0)
        data = self._fp.read()
        self._fp.seek(pos)
        return data

    def dump(self, path):
        """Write test file data to the filesystem.

        Args:
            path (str): Path to output data

        Returns:
            None
        """
        target_path = os.path.join(path, os.path.dirname(self.file_name))
        if not os.path.isdir(target_path):
            os.makedirs(target_path)
        self._fp.seek(0)
        with open(os.path.join(path, self.file_name), "wb") as dst_fp:
            shutil.copyfileobj(self._fp, dst_fp, self.XFER_BUF)

    @classmethod
    def from_data(cls, data, file_name, encoding="UTF-8"):
        """Create a TestFile and add it to the test case.

        Args:
            data (bytes): Data to write to file
            file_name (str): Name for the test file
            encoding (str): Encoding to be used

        Returns:
            TestFile: new instance
        """
        t_file = cls(file_name)
        if data:
            if isinstance(data, bytes) or not encoding:
                t_file.write(data)
            else:
                t_file.write(data.encode(encoding))
        return t_file

    @classmethod
    def from_file(cls, input_file, file_name=None):
        """Create a TestFile from an existing file.

        Args:
            input_file (str): Path to existing file to use
            file_name (str): Name for the test file

        Returns:
            TestFile: new instance
        """
        if file_name is None:
            file_name = os.path.basename(input_file)
        t_file = cls(file_name)
        with open(input_file, "rb") as src_fp:
            shutil.copyfileobj(src_fp, t_file._fp, cls.XFER_BUF)  # pylint: disable=protected-access
        return t_file

    @property
    def size(self):
        """Size of the file in bytes.

        Args:
            None

        Returns:
            int: Size in bytes.
        """
        pos = self._fp.tell()
        self._fp.seek(0, os.SEEK_END)
        size = self._fp.tell()
        self._fp.seek(pos)
        return size

    def write(self, data):
        """Add data to the TestFile.

        Args:
            data (bytes): Data to add to the TestFile

        Returns:
            None
        """
        self._fp.write(data)



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from logging import getLogger
from os import close, getenv, kill, makedirs, unlink
from os.path import isdir, isfile
from platform import system
import signal
from time import localtime, sleep, strftime, time
from tempfile import mkdtemp, mkstemp

from psutil import AccessDenied, NoSuchProcess, Process, process_iter

from ffpuppet import BrowserTimeoutError, FFPuppet, LaunchError
from prefpicker import PrefPicker

from .target_monitor import TargetMonitor
from .target import Target, TargetLaunchError, TargetLaunchTimeout, TargetError
from ..common.utils import grz_tmp


__all__ = ("PuppetTarget",)
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith", "Jesse Schwartzentruber"]

LOG = getLogger("puppet_target")


class PuppetTarget(Target):
    __slots__ = ("use_rr", "use_valgrind", "_browser_logs", "_puppet", "_tmp_prefs")

    def __init__(self, binary, extension, launch_timeout, log_limit, memory_limit, prefs, relaunch, **kwds):
        super(PuppetTarget, self).__init__(binary, extension, launch_timeout, log_limit,
                                           memory_limit, prefs, relaunch)
        self.use_rr = kwds.pop("rr", False)
        self.use_valgrind = kwds.pop("valgrind", False)
        use_xvfb = kwds.pop("xvfb", False)
        if kwds:
            LOG.warning("PuppetTarget ignoring unsupported arguments: %s", ", ".join(kwds))
        self._browser_logs = None
        # generate prefs.js file if needed
        if self.prefs is None:
            for prefs_template in PrefPicker.templates():
                if prefs_template.endswith("browser-fuzzing.yml"):
                    LOG.debug("using prefpicker template %r", prefs_template)
                    tmp_fd, self.prefs = mkstemp(prefix="prefs_", suffix=".js", dir=grz_tmp())
                    close(tmp_fd)
                    PrefPicker.load_template(prefs_template).create_prefsjs(self.prefs)
                    LOG.debug("generated prefs.js %r", self.prefs)
                    break
            else:  # pragma: no cover
                raise TargetError("Failed to generate prefs.js")
            self._tmp_prefs = True
        else:
            self._tmp_prefs = False
        # create Puppet object
        self._puppet = FFPuppet(
            use_rr=self.use_rr,
            use_valgrind=self.use_valgrind,
            use_xvfb=use_xvfb)

    def _abort_hung_proc(self):
        # send SIGABRT to the busiest process
        with self._lock:
            proc_usage = self._puppet.cpu_usage()
        for pid, cpu in sorted(proc_usage, reverse=True, key=lambda x: x[1]):
            LOG.debug("sending SIGABRT to pid: %r, cpu: %0.2f%%", pid, cpu)
            kill(pid, signal.SIGABRT)
            break

    def add_abort_token(self, token):
        self._puppet.add_abort_token(token)

    def cleanup(self):
        # prevent parallel calls to FFPuppet.close() and/or FFPuppet.clean_up()
        if self._browser_logs:
            self.close()
        with self._lock:
            self._puppet.clean_up()
        if self._tmp_prefs and isfile(self.prefs):
            unlink(self.prefs)

    def close(self):
        # prevent parallel calls to FFPuppet.close() and/or FFPuppet.clean_up()
        with self._lock:
            self._puppet.close()
            # save logs in lock to avoid a parallel clean_up() removing them
            if self._browser_logs:
                log_path = mkdtemp(
                    prefix=strftime("%Y%m%d-%H%M%S_", localtime()),
                    suffix="_browser_logs",
                    dir=self._browser_logs)
                LOG.debug("saving browser logs to %r", log_path)
                self._puppet.save_logs(log_path)
                # only save logs once per launch
                self._browser_logs = None

    @property
    def closed(self):
        return self._puppet.reason is not None

    def is_idle(self, threshold):
        for _, cpu in self._puppet.cpu_usage():
            if cpu >= threshold:
                return False
        return True

    @property
    def monitor(self):
        if self._monitor is None:
            class _PuppetMonitor(TargetMonitor):
                # pylint: disable=no-self-argument,protected-access
                def clone_log(_, log_id, offset=0):
                    return self._puppet.clone_log(log_id, offset=offset)
                def is_running(_):
                    return self._puppet.is_running()
                def is_healthy(_):
                    return self._puppet.is_healthy()
                @property
                def launches(_):
                    return self._puppet.launches
                def log_length(_, log_id):
                    return self._puppet.log_length(log_id)
            self._monitor = _PuppetMonitor()
        return self._monitor

    def detect_failure(self, ignored, was_timeout):
        status = self.RESULT_NONE
        if self.expect_close and not was_timeout:
            # give the browser a moment to close if needed
            self._puppet.wait(timeout=30)
        is_healthy = self._puppet.is_healthy()
        # check if there has been a crash, hang, etc...
        if not is_healthy or was_timeout:
            if self._puppet.is_running():
                LOG.debug("terminating browser...")
                if was_timeout and "timeout" not in ignored and system() == "Linux":
                    self._abort_hung_proc()
                    # give the process a moment to start dump
                    self._puppet.wait(timeout=1)
            self.close()
        # if something has happened figure out what
        if not is_healthy:
            if self._puppet.reason == FFPuppet.RC_CLOSED:
                LOG.info("target.close() was called")
            elif self._puppet.reason == FFPuppet.RC_EXITED:
                LOG.info("Target closed itself")
            elif (self._puppet.reason == FFPuppet.RC_WORKER
                  and "memory" in ignored
                  and "ffp_worker_memory_usage" in self._puppet.available_logs()):
                status = self.RESULT_IGNORED
                LOG.info("Memory limit exceeded")
            elif (self._puppet.reason == FFPuppet.RC_WORKER
                  and "log-limit" in ignored
                  and "ffp_worker_log_size" in self._puppet.available_logs()):
                status = self.RESULT_IGNORED
                LOG.info("Log size limit exceeded")
            else:
                LOG.debug("failure detected, ffpuppet return code: %r", self._puppet.reason)
                status = self.RESULT_FAILURE
        elif was_timeout:
            LOG.info("Timeout detected")
            status = self.RESULT_IGNORED if "timeout" in ignored else self.RESULT_FAILURE
        return status

    def dump_coverage(self, timeout=15):
        pid = self._puppet.get_pid()
        if pid is None or not self._puppet.is_healthy():
            LOG.debug("Skipping coverage dump (target is not in a good state)")
            return
        # If at this point, the browser is in a good state, i.e. no crashes
        # or hangs, so signal the browser to dump coverage.
        try:
            for child in Process(pid).children(recursive=True):
                LOG.debug("Sending SIGUSR1 to %d (child)", child.pid)
                kill(child.pid, signal.SIGUSR1)
        except (AccessDenied, NoSuchProcess):  # pragma: no cover
            pass
        LOG.debug("Sending SIGUSR1 to %d (parent)", pid)
        kill(pid, signal.SIGUSR1)
        start_time = time()
        gcda_found = False
        delay = 0.1
        # wait for processes to write .gcno files
        # this should typically take less than 1 second
        while True:
            for proc in process_iter(attrs=["pid", "ppid", "open_files"]):
                # check if proc is the target or child process
                if pid in (proc.info["pid"], proc.info["ppid"]):
                    if proc.info["open_files"] is None:
                        continue
                    if any(x.path.endswith(".gcda") for x in proc.info["open_files"]):
                        gcda_found = True
                        # get the pid of the process that has the file open
                        gcda_open = proc.info["pid"]
                        break
            else:
                gcda_open = None
            elapsed = time() - start_time
            if gcda_found:
                if gcda_open is None:
                    # success
                    LOG.debug("gcda dump took %0.2fs", elapsed)
                    break
                if elapsed >= timeout:
                    # timeout failure
                    LOG.warning("gcda file open by pid %d after %0.2fs", gcda_open, elapsed)
                    kill(gcda_open, signal.SIGABRT)
                    sleep(1)
                    self.close()
                    break
                if delay < 1.0:
                    # increase delay to a maximum of 1 second
                    delay = min(1.0, delay + 0.1)
            elif elapsed >= 3:
                # assume we missed the process writing .gcno files
                LOG.warning("No gcda files seen after %0.2fs", elapsed)
                break
            if not self._puppet.is_healthy():
                LOG.warning("Browser failure during dump_coverage()")
                break
            sleep(delay)

    def launch(self, location, env_mod=None):
        # GRZ_BROWSER_LOGS is intended to be used to aid in debugging.
        # when close() is called a copy of the browser logs will be saved
        # to the directory specified by GRZ_BROWSER_LOGS
        self._browser_logs = getenv("GRZ_BROWSER_LOGS")
        if self._browser_logs and not isdir(self._browser_logs):
            makedirs(self._browser_logs)
        self.rl_countdown = self.rl_reset
        env_mod = dict(env_mod or [])
        # do not allow network connections to non local endpoints
        env_mod["MOZ_DISABLE_NONLOCAL_CONNECTIONS"] = "1"
        env_mod["MOZ_CRASHREPORTER_SHUTDOWN"] = "1"
        try:
            self._puppet.launch(
                self.binary,
                launch_timeout=self.launch_timeout,
                location=location,
                log_limit=self.log_limit,
                memory_limit=self.memory_limit,
                prefs_js=self.prefs,
                extension=self.extension,
                env_mod=env_mod)
        except LaunchError as exc:
            LOG.error("FFPuppet LaunchError: %s", str(exc))
            self.close()
            if isinstance(exc, BrowserTimeoutError):
                raise TargetLaunchTimeout(str(exc))
            raise TargetLaunchError(str(exc))

    def log_size(self):
        return self._puppet.log_length("stderr") + self._puppet.log_length("stdout")

    def save_logs(self, *args, **kwargs):
        self._puppet.save_logs(*args, **kwargs)



# coding=utf-8
"""
Sapphire HTTP server
"""
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import errno
import logging
import os
import random
import shutil
import socket
import tempfile
import time

from .sapphire_job import SapphireJob
from .sapphire_load_manager import SapphireLoadManager
from .status_codes import SERVED_ALL, SERVED_NONE, SERVED_TIMEOUT


__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]


LOG = logging.getLogger("sapphire")


class Sapphire(object):
    def __init__(self, allow_remote=False, auto_close=-1, max_workers=10, port=None, timeout=60):
        self._auto_close = auto_close  # call 'window.close()' on 4xx error pages
        self._max_workers = max_workers  # limit worker threads
        self._socket = Sapphire._create_listening_socket(allow_remote, port)
        self._timeout = None
        self.timeout = timeout

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.close()

    @staticmethod
    def _create_listening_socket(allow_remote, requested_port, retries=20):
        # The intention of this function is to contain the socket creation code
        # along with all the searching and retrying code. If a specific port is requested
        # and it is not available a socket.error will be raised.
        for retry in reversed(range(retries)):
            sock = None
            try:
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                sock.settimeout(0.25)
                # find an unused port and avoid blocked ports
                # see: dxr.mozilla.org/mozilla-central/source/netwerk/base/nsIOService.cpp
                port = random.randint(0x2000, 0xFFFF) if requested_port is None else requested_port
                sock.bind(("0.0.0.0" if allow_remote else "127.0.0.1", port))
                sock.listen(5)
            except (OSError, socket.error) as soc_e:
                if sock is not None:
                    sock.close()
                if retry > 1 and soc_e.errno in (errno.EADDRINUSE, 10013):
                    time.sleep(0.1)
                    continue
                raise
            break
        return sock

    def close(self):
        """
        close()

        This function closes the listening server socket if it is open.
        """
        if self._socket is not None:
            self._socket.close()

    @property
    def port(self):
        """
        port -> int

        returns the port number the socket is listening on
        """

        return self._socket.getsockname()[1]

    def serve_path(self, path, continue_cb=None, forever=False, optional_files=None, server_map=None):
        """
        serve_path() -> tuple
        path is the directory that will be used as wwwroot. The callback continue_cb should
        be a function that returns True or False. If continue_cb is specified and returns False
        the server serve loop will exit. optional_files is list of files that do not need to be
        served in order to exit the serve loop.

        returns a tuple (server status, files served)
        server status is an int:
        - SERVED_ALL: All files excluding files int the optional_files list were served
        - SERVED_NONE: No files were served
        - SERVED_REQUEST: Some files were requested
        files served is a list of the files that were served
        """
        LOG.debug("serving %r (forever=%r)", path, forever)
        job = SapphireJob(
            path,
            auto_close=self._auto_close,
            forever=forever,
            optional_files=optional_files,
            server_map=server_map)
        if not job.pending:
            job.finish()
            LOG.debug("nothing to serve")
            return (SERVED_NONE, tuple())
        with SapphireLoadManager(job, self._socket, self._max_workers) as loadmgr:
            was_timeout = not loadmgr.wait(self.timeout, continue_cb=continue_cb)
        LOG.debug("status: %r, timeout: %r", job.status, was_timeout)
        return (SERVED_TIMEOUT if was_timeout else job.status, tuple(job.served))

    def serve_testcase(self, testcase, continue_cb=None, forever=False, working_path=None, server_map=None):
        """
        serve_testcase() -> tuple
        testcase is the Grizzly TestCase to serve. The callback continue_cb should
        be a function that returns True or False. If continue_cb is specified and returns False
        the server serve loop will exit. working_path is where the testcase will be unpacked
        temporary.

        returns a tuple (server status, files served)
        see serve_path() for more info
        """
        LOG.debug("serve_testcase() called")
        wwwdir = tempfile.mkdtemp(prefix="sphr_test_", dir=working_path)
        try:
            testcase.dump(wwwdir)
            serve_start = time.time()
            result = self.serve_path(
                wwwdir,
                continue_cb=continue_cb,
                forever=forever,
                optional_files=tuple(testcase.optional),
                server_map=server_map)
            testcase.duration = time.time() - serve_start
        finally:
            # remove test case working directory
            shutil.rmtree(wwwdir, ignore_errors=True)
        return result

    @property
    def timeout(self):
        return self._timeout

    @timeout.setter
    def timeout(self, value):
        if not value:
            self._timeout = 0
        else:
            self._timeout = max(value, 1)

    @classmethod
    def main(cls, args):
        try:
            with cls(allow_remote=args.remote, port=args.port, timeout=args.timeout) as serv:
                LOG.info(
                    "Serving %r @ http://%s:%d/",
                    os.path.abspath(args.path),
                    socket.gethostname() if args.remote else "127.0.0.1",
                    serv.port)
                status = serv.serve_path(args.path)[0]
            if status == SERVED_ALL:
                LOG.info("All test case content was served")
            else:
                LOG.warning("Failed to serve all test content")
        except KeyboardInterrupt:
            LOG.warning("Ctrl+C detected. Shutting down...")



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import pytest

from .adapter import Adapter, AdapterError


class SimpleAdapter(Adapter):
    NAME = "simple"

    def generate(self, testcase, server_map):
        pass


def test_adapter_01():
    """test a bad Adapter"""
    class BadAdapter(SimpleAdapter):
        NAME = None
    with pytest.raises(AdapterError, match="BadAdapter.NAME must be a string"):
        BadAdapter()

def test_adapter_02():
    """test a simple Adapter"""
    adpt = SimpleAdapter()
    assert isinstance(adpt.fuzz, dict)
    assert not adpt.fuzz
    assert adpt.monitor is None
    assert adpt.remaining is None
    assert adpt.get_harness() is None
    adpt.setup(None, None)
    adpt.generate(None, None)
    adpt.on_served(None, None)
    adpt.on_timeout(None, None)
    adpt.pre_launch()
    adpt.cleanup()

def test_adapter_03(tmp_path):
    """test Adapter.enable_harness()"""
    adpt = SimpleAdapter()
    # built-in harness
    harness_file = tmp_path / "harness.html"
    test_data = b"default_harness_data"
    harness_file.write_bytes(test_data)
    adpt.HARNESS_FILE = str(harness_file)
    adpt.enable_harness()
    assert adpt.get_harness() == test_data
    # external harness
    ext_harness_file = tmp_path / "ext_harness.html"
    test_data = b"external_harness_data"
    ext_harness_file.write_bytes(test_data)
    adpt.enable_harness(str(ext_harness_file))
    assert adpt.get_harness() == test_data

def test_adapter_04(tmp_path):
    """test Adapter.scan_path()"""
    # empty path
    assert not any(SimpleAdapter.scan_path(str(tmp_path)))
    # missing path
    assert not any(SimpleAdapter.scan_path(str(tmp_path / "none")))
    # path to file
    file1 = (tmp_path / "test1.txt")
    file1.touch()
    found = tuple(SimpleAdapter.scan_path(str(file1)))
    assert str(file1) in found
    assert len(found) == 1
    # path to directory
    assert len(tuple(SimpleAdapter.scan_path(str(tmp_path)))) == 1
    # path to directory (w/ ignored)
    (tmp_path / ".ignored").touch()
    nested = (tmp_path / "nested")
    nested.mkdir()
    file2 = (nested / "test2.bin")
    file2.touch()
    assert len(tuple(SimpleAdapter.scan_path(str(tmp_path)))) == 1
    # path to directory (recursive)
    found = tuple(SimpleAdapter.scan_path(str(tmp_path), recursive=True))
    assert str(file1) in found
    assert str(file2) in found
    assert len(found) == 2



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
Sapphire HTTP server job
"""

from collections import defaultdict, namedtuple
from logging import getLogger
import os
from queue import Queue
import threading

from .server_map import Resource
from .status_codes import SERVED_ALL, SERVED_NONE, SERVED_REQUEST

__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

LOG = getLogger("sphr_job")


Tracker = namedtuple("Tracker", "files lock")


class SapphireJob(object):
    __slots__ = (
        "_complete", "_pending", "_served", "auto_close", "accepting", "base_path",
        "exceptions", "forever", "initial_queue_size", "server_map", "worker_complete")

    def __init__(self, base_path, auto_close=-1, forever=False, optional_files=None, server_map=None):
        self._complete = threading.Event()
        self._pending = Tracker(files=set(), lock=threading.Lock())
        self._served = Tracker(files=defaultdict(int), lock=threading.Lock())
        self.accepting = threading.Event()
        self.accepting.set()
        self.auto_close = auto_close
        self.base_path = os.path.abspath(base_path)  # wwwroot
        self.exceptions = Queue()
        self.forever = forever
        self.initial_queue_size = 0
        self.server_map = server_map
        self.worker_complete = threading.Event()
        self._build_queue(optional_files)

    def _build_queue(self, optional_files):
        # build file list to track files that must be served
        # this is intended to only be called once by __init__()
        for d_name, _, filenames in os.walk(self.base_path, followlinks=False):
            for f_name in filenames:
                # do not add optional files to queue of required files
                if optional_files and f_name in optional_files:
                    LOG.debug("optional: %r", f_name)
                    continue
                file_path = os.path.abspath(os.path.join(d_name, f_name))
                if "?" in file_path:
                    LOG.warning("Cannot add files with '?' in path. Skipping %r", file_path)
                    continue
                self._pending.files.add(file_path)
                LOG.debug("required: %r", f_name)
        # if nothing was found check if the path exists
        if not self._pending.files and not os.path.isdir(self.base_path):
            raise OSError("%r does not exist" % (self.base_path),)
        if self.server_map:
            for redirect, resource in self.server_map.redirect.items():
                if resource.required:
                    self._pending.files.add(redirect)
                LOG.debug(
                    "%s: %r -> %r",
                    "required" if resource.required else "optional",
                    redirect,
                    resource.target)
        self.initial_queue_size = len(self._pending.files)
        LOG.debug("%d files required to serve", self.initial_queue_size)

    def check_request(self, request):
        if "?" in request:
            request = request.split("?", 1)[0]
        to_serve = os.path.normpath(os.path.join(self.base_path, request))
        if "\x00" not in to_serve and os.path.isfile(to_serve):
            res = Resource(Resource.URL_FILE, to_serve)
            with self._pending.lock:
                res.required = to_serve in self._pending.files
            return res
        if self.server_map is not None:
            if request in self.server_map.redirect:
                return self.server_map.redirect[request]
            if request in self.server_map.dynamic:
                return self.server_map.dynamic[request]
            # collect possible include matches
            includes = tuple(x for x in self.server_map.include if request.startswith(x))
            if includes:
                LOG.debug("potential include matches %r", includes)
                # attempt to find match
                url = request
                while True:
                    if url in includes:
                        LOG.debug("found include match %r", url)
                        location = request.split(url, 1)[-1].lstrip("/") if url else request
                        # check location points to something
                        if location:
                            target = os.path.join(
                                self.server_map.include[url].target,
                                location)
                            # if the mapping url is empty check the file exists
                            if url or os.path.isfile(target):
                                return Resource(
                                    Resource.URL_INCLUDE,
                                    os.path.normpath(target),
                                    mime=self.server_map.include[url].mime,
                                    required=self.server_map.include[url].required)
                    if "/" in url:
                        url = url.rsplit("/", 1)[0]
                    elif url:
                        # try empty mount point
                        url = ""
                    else:
                        # include does not exist
                        break
        return None

    def finish(self):
        self._complete.set()

    def increment_served(self, target):
        # update list of served files
        with self._served.lock:
            self._served.files[target] += 1

    def is_complete(self, wait=None):
        if wait is not None:
            return self._complete.wait(wait)
        return self._complete.is_set()

    def is_forbidden(self, target_file):
        target_file = os.path.abspath(target_file)
        # check if target_file lives somewhere in wwwroot
        if not target_file.startswith(self.base_path):
            if self.server_map:
                for resources in self.server_map.include.values():
                    if target_file.startswith(resources.target):
                        return False  # this is a valid include path
            return True  # this is NOT a valid include path
        return False  # this is a valid path

    @property
    def pending(self):
        # number of pending files
        with self._pending.lock:
            return len(self._pending.files)

    def remove_pending(self, file_name):
        # return True when all file have been removed
        with self._pending.lock:
            if self._pending.files:
                self._pending.files.discard(file_name)
            return not self._pending.files

    @property
    def served(self):
        # served files
        # files served from www root will have a path relative to www root
        # include files will have an absolute path
        with self._served.lock:
            # make a copy of what is available (maybe a copy not necessary?)
            served = tuple(self._served.files.keys())
        for fname in served:
            if fname.startswith(self.base_path):
                # file is in www root
                yield os.path.relpath(fname, self.base_path)
            else:
                # include file
                yield fname

    @property
    def status(self):
        with self._pending.lock:
            queue_size = len(self._pending.files)
        if queue_size == 0:
            return SERVED_ALL
        if queue_size < self.initial_queue_size:
            return SERVED_REQUEST
        return SERVED_NONE



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import collections
import json
import logging
import os
import sys
import tempfile

from Collector.Collector import Collector

from .args import ReducerFuzzManagerIDQualityArgs
from .crash import CrashReductionJob


LOG = logging.getLogger("grizzly.reduce.bucket")


def bucket_crashes(bucket_id, quality_filter):
    """Fetch all crash IDs for the specified FuzzManager bucket.
    Only crashes with testcases are returned.

    Args:
        bucket_id (int): ID of the requested bucket on the server side
        quality_filter (int): Filter crashes by quality value (None for all)

    Returns:
        generator: generator of crash ID (int)
    """
    coll = Collector()

    def _get_results(endpoint, params=None):
        """
        Function to get paginated results from FuzzManager

        Args:
            endpoint (str): FuzzManager REST API to query (eg. "crashes").
            params (dict): Params to pass through to requests.get

        Returns:
            generator: objects returned by FuzzManager (as dicts)
        """
        LOG.debug("first request to /%s/", endpoint)

        url = "%s://%s:%d/crashmanager/rest/%s/" \
            % (coll.serverProtocol, coll.serverHost, coll.serverPort, endpoint)

        response = coll.get(url, params=params).json()

        while True:
            LOG.debug("got %d/%d %s", len(response["results"]), response["count"], endpoint)
            while response["results"]:
                yield response["results"].pop()

            if response["next"] is None:
                break

            LOG.debug("next request to /%s/", endpoint)
            response = coll.get(response["next"]).json()

    # Get all crashes for bucket
    query_args = [
        ("op", "AND"),
        ("bucket", bucket_id),
    ]
    if quality_filter is not None:
        query_args.append(("testcase__quality", quality_filter))
    query = json.dumps(collections.OrderedDict(query_args))

    n_yielded = 0
    for crash in _get_results("crashes", params={"query": query, "include_raw": "0"}):

        if not crash["testcase"]:
            LOG.warning("crash %d has no testcase, skipping", crash["id"])
            continue

        n_yielded += 1
        LOG.debug("yielding crash #%d", n_yielded)
        yield crash["id"]


def get_signature(bucket_id):
    """
    Download the signature for the specified FuzzManager bucket.

    Args:
        bucket_id (int): ID of the requested bucket on the server side

    Returns:
        str: temp filename to the JSON signature. caller must remove filename when done
    """
    coll = Collector()

    url = "%s://%s:%d/crashmanager/rest/buckets/%d/" \
        % (coll.serverProtocol, coll.serverHost, coll.serverPort, bucket_id)

    response = coll.get(url).json()

    sig_fd, sig_fn = tempfile.mkstemp(suffix=".json")
    with os.fdopen(sig_fd, "w") as sig_fp:
        sig_fp.write(response["signature"])

    return sig_fn


def main(args):
    LOG.info("Trying all crashes in bucket %d until one reduces", args.input)

    # if no signature specified, download the signature from FM
    rm_sig = False
    if not args.sig:
        args.sig = get_signature(args.input)
        rm_sig = True

    try:
        for crash_id in bucket_crashes(args.input, args.quality):

            # reduce.main expects input to be a crash ID
            args.input = crash_id

            if CrashReductionJob.main(args) == 0:
                # success!
                return 0

        # none of the testcases reduced
        return 1

    finally:
        if rm_sig:
            os.unlink(args.sig)


if __name__ == "__main__":
    sys.exit(main(ReducerFuzzManagerIDQualityArgs().parse_args()))



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import logging
import os
import re


__all__ = ("Resource", "ServerMap")
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

LOG = logging.getLogger("sphr_map")  # pylint: disable=invalid-name


class InvalidURLError(Exception):
    """Raised what a URL string string contained invalid characters"""


class MapCollisionError(Exception):
    """Raised when a URL is already in use by ServerMap"""


class Resource(object):
    URL_DYNAMIC = 0
    URL_FILE = 1
    URL_INCLUDE = 2
    URL_REDIRECT = 3

    __slots__ = ("mime", "required", "target", "type")

    def __init__(self, resource_type, target, mime=None, required=False):
        self.mime = mime
        self.required = required
        self.target = target
        self.type = resource_type


class ServerMap(object):
    __slots__ = ("dynamic", "include", "redirect")

    def __init__(self):
        self.dynamic = dict()
        self.include = dict()  # mapping of directories that can be requested
        self.redirect = dict()  # document paths to map to file names using 307s

    @staticmethod
    def _check_url(url):
        # check and sanitize URL
        url = url.strip("/")
        if re.search(r"\W", url) is not None:
            raise InvalidURLError("Only alpha-numeric characters accepted in URL.")
        return url

    def set_dynamic_response(self, url, callback, mime_type="application/octet-stream"):
        url = self._check_url(url)
        if not callable(callback):
            raise TypeError("callback must be callable")
        if not isinstance(mime_type, str):
            raise TypeError("mime_type must be of type 'str'")
        if url in self.include or url in self.redirect:
            raise MapCollisionError("URL collision on %r" % (url,))
        LOG.debug("mapping dynamic response %r -> %r (%r)", url, callback, mime_type)
        self.dynamic[url] = Resource(
            Resource.URL_DYNAMIC,
            callback,
            mime=mime_type)

    def set_include(self, url, target_path):
        url = self._check_url(url)
        if not os.path.isdir(target_path):
            raise IOError("Include path not found: %s" % (target_path,))
        if url in self.dynamic or url in self.redirect:
            raise MapCollisionError("URL collision on %r" % (url,))
        target_path = os.path.abspath(target_path)
        # sanity check to prevent mapping overlapping paths
        # Note: This was added to help map file served via includes back to
        # the files on disk. This is a temporary workaround until mapping of
        # requests to files that were served is available outside of Sapphire.
        for existing_url, resource in self.include.items():
            if url == existing_url:
                # allow overwriting entry
                continue
            if not os.path.relpath(target_path, resource.target).startswith(".."):
                LOG.error("%r mapping includes path %r", existing_url, target_path)
                raise MapCollisionError("%r and %r include %r" % (url, existing_url, target_path))
            if not os.path.relpath(resource.target, target_path).startswith(".."):
                LOG.error("%r mapping includes path %r", url, resource.target)
                raise MapCollisionError("%r and %r include %r" % (url, existing_url, resource.target))
        LOG.debug("mapping include %r -> %r", url, target_path)
        self.include[url] = Resource(
            Resource.URL_INCLUDE,
            target_path)

    def set_redirect(self, url, target, required=True):
        url = self._check_url(url)
        if not isinstance(target, str):
            raise TypeError("target must be of type 'str'")
        if not target:
            raise TypeError("target must not be an empty string")
        if url in self.dynamic or url in self.include:
            raise MapCollisionError("URL collision on %r" % (url,))
        self.redirect[url] = Resource(
            Resource.URL_REDIRECT,
            target,
            required=required)



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from logging import getLogger
from sys import exc_info
from traceback import extract_tb
from pkg_resources import iter_entry_points

from .target import sanitizer_opts, Target, TargetError, TargetLaunchError, TargetLaunchTimeout

__all__ = ("Target", "TargetError", "TargetLaunchError", "TargetLaunchTimeout",
           "available", "load", "sanitizer_opts")
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith", "Jesse Schwartzentruber"]

TARGETS = None
LOG = getLogger("grizzly")


def _load_targets():
    global TARGETS  # pylint: disable=global-statement
    TARGETS = {}
    for entry_point in iter_entry_points('grizzly_targets'):
        LOG.debug("scanning target %r", entry_point.name)
        try:
            target = entry_point.load()
        except Exception:  # pylint: disable=broad-except
            exc_type, exc_obj, exc_tb = exc_info()
            tbinfo = extract_tb(exc_tb)[-1]
            LOG.warning("Target %r raised an exception %s: %s (%s:%d)", entry_point.name, exc_type.__name__,
                        exc_obj, tbinfo[0], tbinfo[1])
            continue
        if not issubclass(target, Target):
            LOG.warning("Target %r doesn't inherit from grizzly.target.Target, skipping.", entry_point.name)
        elif entry_point.name in TARGETS:
            raise RuntimeError("Target %r already exists as %r. (duplicate: %r)" %
                               (entry_point.name, TARGETS[entry_point.name], target))
        else:
            TARGETS[entry_point.name] = target


def available():
    if TARGETS is None:
        _load_targets()
    return TARGETS.keys()


def load(name):
    if TARGETS is None:
        _load_targets()
    return TARGETS[name]



# coding=utf-8
"""
SapphireLoadManager unit tests
"""
# pylint: disable=protected-access

import socket
import threading

import pytest

from .sapphire_load_manager import SapphireLoadManager
from .sapphire_job import SapphireJob


def test_sapphire_load_manager_01(mocker, tmp_path):
    """test basic SapphireLoadManager"""
    (tmp_path / "testfile").write_bytes(b"test")
    job = SapphireJob(str(tmp_path))
    clnt_sock = mocker.Mock(spec=socket.socket)
    clnt_sock.recv.return_value = b"GET /testfile HTTP/1.1"
    serv_sock = mocker.Mock(spec=socket.socket)
    serv_sock.accept.return_value = (clnt_sock, None)
    assert not job.is_complete()
    with SapphireLoadManager(job, serv_sock) as loadmgr:
        assert loadmgr.wait(1)
    assert clnt_sock.close.call_count == 1
    assert job.is_complete()
    assert not job.accepting.is_set()
    assert job.exceptions.empty()

def test_sapphire_load_manager_02(mocker):
    """test SapphireLoadManager.start() failure"""
    mocker.patch("sapphire.sapphire_load_manager.time.sleep", autospec=True)
    fake_thread = mocker.patch("sapphire.sapphire_load_manager.threading.Thread", autospec=True)
    fake_thread.return_value.start.side_effect = threading.ThreadError
    job = mocker.Mock(spec=SapphireJob)
    job.pending = True
    loadmgr = SapphireLoadManager(job, None)
    with pytest.raises(threading.ThreadError):
        loadmgr.start()
    loadmgr.close()
    assert job.is_complete()

def test_sapphire_load_manager_03(mocker, tmp_path):
    """test SapphireLoadManager multiple files and requests"""
    (tmp_path / "test1").touch()
    (tmp_path / "test2").touch()
    (tmp_path / "test3").touch()
    job = SapphireJob(str(tmp_path))
    clnt_sock = mocker.Mock(spec=socket.socket)
    clnt_sock.recv.side_effect = (
        b"GET /test1 HTTP/1.1",
        b"GET /missing HTTP/1.1",
        b"badrequest",
        b"",
        b"GET /test2 HTTP/1.1",
        b"GET /test1 HTTP/1.1",
        b"GET /test1 HTTP/1.1",
        b"GET /test3 HTTP/1.1")
    serv_sock = mocker.Mock(spec=socket.socket)
    serv_sock.accept.return_value = (clnt_sock, None)
    assert not job.is_complete()
    with SapphireLoadManager(job, serv_sock, max_workers=2) as loadmgr:
        assert loadmgr.wait(1)
    assert clnt_sock.close.call_count == 8
    assert job.is_complete()

def test_sapphire_load_manager_04(mocker, tmp_path):
    """test SapphireLoadManager.wait()"""
    (tmp_path / "test1").touch()
    job = SapphireJob(str(tmp_path))
    clnt_sock = mocker.Mock(spec=socket.socket)
    clnt_sock.recv.return_value = b""
    serv_sock = mocker.Mock(spec=socket.socket)
    serv_sock.accept.return_value = (clnt_sock, None)
    with SapphireLoadManager(job, serv_sock, max_workers=10) as loadmgr:
        # invalid callback
        with pytest.raises(TypeError, match="continue_cb must be callable"):
            loadmgr.wait(0, continue_cb="test")
        # callback abort
        assert loadmgr.wait(1, continue_cb=lambda: False, poll=0.01)
    # timeout
    job = SapphireJob(str(tmp_path))
    fake_time = mocker.patch("sapphire.sapphire_load_manager.time", autospec=True)
    fake_time.time.side_effect = (1, 2, 3)
    with SapphireLoadManager(job, serv_sock, max_workers=10) as loadmgr:
        assert not loadmgr.wait(1, continue_cb=lambda: False, poll=0.01)

def test_sapphire_load_manager_05(mocker, tmp_path):
    """test SapphireLoadManager re-raise worker exceptions"""
    (tmp_path / "test1").touch()
    job = SapphireJob(str(tmp_path))
    clnt_sock = mocker.Mock(spec=socket.socket)
    clnt_sock.recv.side_effect = Exception("worker exception")
    serv_sock = mocker.Mock(spec=socket.socket)
    serv_sock.accept.return_value = (clnt_sock, None)
    with pytest.raises(Exception, match="worker exception"):
        with SapphireLoadManager(job, serv_sock) as loadmgr:
            loadmgr.wait(1)
    assert clnt_sock.close.call_count == 1
    assert job.is_complete()
    assert job.exceptions.empty()

def test_sapphire_load_manager_06(mocker, tmp_path):
    """test SapphireLoadManager re-raise launcher exceptions"""
    (tmp_path / "test1").touch()
    job = SapphireJob(str(tmp_path))
    serv_sock = mocker.Mock(spec=socket.socket)
    serv_sock.accept.side_effect = Exception("launcher exception")
    with pytest.raises(Exception, match="launcher exception"):
        with SapphireLoadManager(job, serv_sock) as loadmgr:
            loadmgr.wait(1)
    assert job.is_complete()
    assert job.exceptions.empty()
